import time

from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

"""
TIME EVOLUTION EXAMPLE 1: Vortex trains

"""

if __name__ == "__main__":

    N = 50

    sq_array = SquareArray(N, N)
    # sq_array.set_inductance_factors(1)
    # x = np.ones(sq_array._Nj())
    # x[np.random.rand(sq_array._Nj()) < 0.3] = 0
    # x = np.random.rand(sq_array._Nj())
    # print(len(x))
    # sq_array.set_inductance_factors(x)
    # print(sq_array._get_mixed_inductance_mask())
    # sq_array.set_capacitance_factors(1)

    f = 0.05
    W = 1
    Is = 0.6 * sq_array.current_base(angle=0)[:, None, None] * np.ones((1, W, 1))
    T = 0.01
    dt = 0.02
    Nt = 10000
    store_stepsize = 25
    ts = np.arange(0, Nt, store_stepsize)

    problem = TimeEvolutionProblem(sq_array, time_step=dt,
                                   time_step_count=Nt, current_sources=Is,
                                   frustration=f, temperature=T,
                                   store_time_steps=ts)
    tic = time.perf_counter()
    out = problem.compute()
    print(time.perf_counter() - tic)
    # out.animate(junction_quantity="Isup", face_quantity="flux", face_quantity_clim=[-1, 2],
    #             figsize=[10, 10])
    # out.animate()
    plt.show()




from pyjjasim import *

import matplotlib.pyplot as plt


"""
Static Example 1: Single vortex

Compute a single vortex at the centre of a square array.

"""

if __name__ == "__main__":
    print("STATIC EXAMPLE1: Single vortex")

    # define arrays
    N = 10
    sq_array = SquareArray(N, N)

    # define problem parameters
    f = 0.01                                                          # frustration
    n = np.zeros(sq_array.face_count(), dtype=int)                    # target vortex configuration
    centre_face_idx = sq_array.locate_faces((N - 1) / 2, (N - 1) / 2) # locating face idx at coordinate x=(N-1)/2, y=(N-1)/2
    n[centre_face_idx] = 1

    # define static problems
    problem = StaticProblem(sq_array, frustration=f, vortex_configuration=n)

    # approximations:
    approx_arctan = problem.approximate(algorithm=0)
    approx_london = problem.approximate(algorithm=1)

    print(f"arctan approx error: {approx_arctan.get_error_kirchhoff_rules()} and {approx_arctan.get_error_winding_rules()}")
    print(f"london approx error: {approx_london.get_error_kirchhoff_rules()} and {approx_london.get_error_winding_rules()}")

    approx_arctan.plot(node_quantity="phi", title="arctan approximation")
    approx_london.plot(node_quantity="phi", title="london approximation")

    config1, status1, info1 = problem.compute(initial_guess=approx_arctan, algorithm=1)
    config2, status2, info2 = problem.compute(initial_guess=approx_london, algorithm=1)

    print(f"config1 error: {config1.get_error_kirchhoff_rules()} and {config1.get_error_winding_rules()}")
    print(f"config2 error: {config2.get_error_kirchhoff_rules()} and {config2.get_error_winding_rules()}")


    print(info1)
    print(info2)
    config1.plot(node_quantity="phi", title="exact solution with actan initial guess")
    config2.plot(node_quantity="phi", title="exact solution with london initial guess")

    def principle_value(x):
        return x - 2 * np.pi * np.round(x / (2 * np.pi))

    theta_difference = config1.get_theta() - config2.get_theta()
    print("||conf1.th - conf2.th||:", scipy.linalg.norm(theta_difference))
    print("||pv(conf1.th - conf2.th)||:", scipy.linalg.norm(principle_value(theta_difference)))

    plt.show()

    # conclusion:
    # - both solutions (i.e. both theta's) are the same modulus 2*pi.
    # - the 2*pi multiples mean both solutions live in a different phase-zone

    # In practice one does not need to manually specify initial conditions.
    # By default the london approximation is used as initial condition. One would just do:
    # problem = StaticProblem(sq_array, frustration=f, vortex_configuration=n)
    # config, status, info = problem.compute()








from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")

"""
EXAMPLE 3: Vortex with inductance: Flux quantization

"""

if __name__ == "__main__":
    # Square array with screening
    N = 14
    array = SquareArray(N, N)

    Llist = [0.001, 0.1, 1, 5]

    f = 1 / (N - 1) ** 2  # frustration
    n = np.zeros(array.face_count(), dtype=int)  # target vortex configuration
    centre_face_idx = array.locate_faces((N - 1) / 2, (N - 1) / 2)  # locating face idx at coordinate x=(N-1)/2, y=(N-1)/2
    n[centre_face_idx] = 1
    for i, L in enumerate(Llist):
        array.set_inductance_factors(L)
        prob = StaticProblem(array, frustration=f, vortex_configuration=n)
        out, _, info = prob.compute()
        out.plot(face_quantity="flux", face_quantity_logarithmic_colors=True,
                 face_quantity_clim=[1E-3, 1], title=f"beta_L={L}", arrow_color=[1, 1, 1])


    plt.show()






"""pyJJASim: Josephson Circuit Simulation in Python """


from pyjjasim.variable_row_array import *
from pyjjasim.embedded_graph import *
from pyjjasim.josephson_circuit import *
from pyjjasim.static_problem import *
from pyjjasim.time_evolution import *
from pyjjasim.circuit_visualize import *







import dataclasses

import numpy as np
import scipy.sparse
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from pyjjasim.variable_row_array import VarRowArray


"""
Embedded Graph Module
"""

class NotSingleComponentError(Exception):
    pass

class NotPlanarEmbeddingError(Exception):
    pass

class NonSimpleError(Exception):
    pass

class SelfLoopError(Exception):
    pass

class NodeNotExistError(Exception):
    pass

class EdgeNotExistError(Exception):
    pass


class EmbeddedGraph:

    """
    Class for embedded 2D graphs. Can be used to construct faces and check planarity.

    Definitions:
     - l-cycle:
        Cycle generated by traversing the embedded graph always taking the leftmost turn at
        every node.
     - face-cycle:
        An l-cycle with positive oriented area (using shoelace formula). If the graph is planar,
        face cycles are the boundaries of faces.

    Requirements:
     - No self-loops
     - Simple (no multigraph)

    Parameters
    ----------
    x, y : (N,) array
        Coordinates of nodes of embedded graph.
    node1, node2 : (E,) int array in range(N)
        endpoint nodes of edges in embedded graph. Nodes are referred to by their index in
        the coordinate arrays.
    require_single_component=False :
        If True, an error is raised if the graph is not single-component.
    require_planar_embedding=False :
        If True, an error is raised if the graph is not a planar embedding.

    Raises
    ------
    NonSimpleError
        If graph is not simple.
    SelfLoopError
        if graph contains self-loops.
    NotSingleComponentError
        if graph has disconnected components and require_single_component=True.
    NotPlanarEmbeddingError
        If graph in not a planar embedding and require_planar_embedding=True.
    """

    def __init__(self, x, y, node1, node2, require_single_component=False,
                 require_planar_embedding=False, _edges_are_sorted=None):
        self.x = np.array(x, dtype=np.double).ravel()
        self.y = np.array(y, dtype=np.double).ravel()
        if len(self.x) != len(self.y):
            raise ValueError("x and y must be same size")
        self.N = len(self.x)
        self.node1, self.node2 = np.array(node1, dtype=int).ravel(), np.array(node2, dtype=int).ravel()
        self._assert_edges_correct_shape()
        self.E = len(self.node1)
        self._assert_edges_contain_existing_nodes()
        self._assert_no_self_loop()

        if _edges_are_sorted is None:
            self.edge_permute = np.arange(self.edge_count())
            self.edge_flip = np.zeros(self.edge_count(), dtype=bool)
            self._sort_edges()
        else:
            self.edge_permute, self.edge_flip = _edges_are_sorted

        self._assert_nonsimple()
        self.edge_v_array = self._assign_edge_v_array()

        # quantities are computed and stored only when a method needs them.
        self.F = None
        self.face_permutation = None
        self.faces_v_array = None
        self.face_edges = None
        self.face_nodes = None
        self.face_lengths = None
        self.determinant = None
        self.areas = None
        self.roll_ids = None
        self.boundary_face_indices = None

        if require_single_component:
            self._assert_single_component()
        if require_planar_embedding:
            self._assert_planar_embedding()

    def get_edges(self):
        """
        Return node endpoints of all edges.

        Returns
        -------
        node1, node2 : (E,) int arrays in range(N)
            Node endpoints of all edges.
        """
        return self._de_sort_edges()

    def get_edge_ids(self, node1, node2, return_direction=False):
        """
        Return indices of edges with given endpoint nodes.

        Parameters
        ----------
        node1, node2 : arrays in range(N)
            Endpoints of edges.
        return_direction=False : bool


        Returns
        -------
        edge_ids : array with same size as node1 in range(E)
            Returns indices of edges.
        edge_directions : bool array with same size as node1
            True if aligned with edge direction

        Raises
        ------
        EdgeNotExistError
            If a queried node-pair does not exist.

        """
        node1, node2 = np.array(node1, dtype=int).ravel(), np.array(node2, dtype=int).ravel()
        mask = node1 < node2
        node1, node2 = np.where(mask, node1, node2), np.where(mask, node2, node1)
        if np.any(node1 > self.node1[-1]):
            raise EdgeNotExistError("queried edge that does not exist")
        starts, ends = self.edge_v_array.row_ranges()
        node1_idx = np.searchsorted(self.node1[starts], node1)
        starts, ends = starts[node1_idx], ends[node1_idx]
        while not np.all(node2 <= self.node2[starts]):
            starts[node2 > self.node2[starts]] += 1
            if not np.all(starts < ends):
                raise EdgeNotExistError("queried edge that does not exist")
        if not np.all((self.node1[starts] == node1) & (self.node2[starts] == node2)):
            raise EdgeNotExistError("queried edge that does not exist")
        if not return_direction:
            return self.edge_permute[starts]
        else:
            e = self.edge_permute[starts]
            return e, self.edge_flip[e] != mask

    def add_nodes(self, x, y):
        """
        Add nodes to graph.

        Parameters
        ----------
        x, y : arrays
            Coordinates of the nodes to be added to the graph

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            New graph with nodes added to it.
        """
        return EmbeddedGraph(np.append(self.x, np.array(x, dtype=np.double).ravel()), np.append(self.y, np.array(y, dtype=np.double).ravel()), self.node1, self.node2, require_single_component=False)

    def add_edges(self, node1, node2, require_single_component=False, require_planar_embedding=False):
        """
        Add edges to graph.

        Parameters
        ----------
        node1, node2 : arrays in range(N)
            Endpoints of edges.
        require_single_component=False :
            If True; raises error if the resulting graph is not single-component.
        require_planar_embedding=False :
            If True; raises error if the resulting graph is not a planar embedding.

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            New graph with edges added to it.

        """
        return EmbeddedGraph(self.x, self.y, np.append(self.node1, np.array(node1, dtype=int).ravel()),
                             np.append(self.node2, np.array(node2, dtype=int).ravel()),
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def add_nodes_and_edges(self, x, y, node1, node2, require_single_component=False,
                            require_planar_embedding=False):
        """
        Add nodes and edges to graph.

        Parameters
        ----------
        x, y : arrays
            coordinates of the nodes to be added to the graph
        node1, node2 : arrays in range(N + x.size)
            endpoints of edges. The i-th new node must be referred to by index N + i.
        require_single_component=False :
            If True; raises error if the resulting graph is not single-component
        require_planar_embedding=False :
            If True; raises error if the resulting graph is not a planar embedding

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            new graph with nodes and edges added to it.

        """
        return EmbeddedGraph(np.append(self.x, np.array(x, dtype=np.double).ravel()),
                             np.append(self.y, np.array(y, dtype=np.double).ravel()),
                             np.append(self.node1, np.array(node1, dtype=int).ravel()),
                             np.append(self.node2, np.array(node2, dtype=int).ravel()),
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def remove_nodes(self, node_ids, require_single_component=False, require_planar_embedding=False):
        """
        Remove nodes from graph.

        Parameters
        ----------
        node_ids : int array in range(N)
            Indices of nodes to be removed.
        require_single_component=False :
            If True; raises error if the resulting graph is not single-component.
        require_planar_embedding=False :
            If True; raises error if the resulting graph is not a planar embedding.

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            New graph with nodes removed from it.
        """
        node_ids = np.array(node_ids, dtype=int).ravel()
        node_map = np.ones(self.node_count(), dtype=int)
        node_map[node_ids] = 0
        node_map = np.cumsum(node_map) - node_map
        node_map[node_ids] = -1
        n1, n2 = node_map[self.node1], node_map[self.node2]
        edge_ids = (n1 >= 0) & (n2 >= 0)
        return EmbeddedGraph(np.delete(self.x, node_ids), np.delete(self.y, node_ids),
                             n1[edge_ids], n2[edge_ids], require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def remove_edges_by_ids(self, edge_ids, require_single_component=False, require_planar_embedding=False):
        """
        Remove edges from graph with edge ids as input.

        Parameters
        ----------
        edge_ids : int array in range(E)
            Indices of nodes to be removed.
        require_single_component=False :
            If True; raises error if the resulting graph is not single-component.
        require_planar_embedding=False :
            If True; raises error if the resulting graph is not a planar embedding.

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            new graph with edges removed from it.
        """
        ids_internal = self.edge_permute_inverse[edge_ids]
        return EmbeddedGraph(self.x, self.y, np.delete(self.node1, ids_internal),
                             np.delete(self.node2, ids_internal), _edges_are_sorted=True,
                             require_single_component=require_single_component,
                             require_planar_embedding=require_planar_embedding)

    def remove_edges(self, node1, node2, require_single_component=False, require_planar_embedding=False):
        """
        Remove edges from graph with node endpoints as input.

        Parameters
        ----------
        node1, node2 : int arrays in range(N)
            The indices of node endpoints of edges to be removed.
        require_single_component=False :
            If True; raises error if the resulting graph is not single-component.
        require_planar_embedding=False :
            If True; raises error if the resulting graph is not a planar embedding.

        Returns
        -------
        new_graph : :py:attr:`EmbeddedGraph`
            new graph with edges removed from it.

        Raises
        ------
        EdgeNotExistError
            If a queried node-pair does not exist.
        """
        return self.remove_edges_by_ids(self.get_edge_ids(node1, node2),
                                        require_single_component=require_single_component,
                                        require_planar_embedding=require_planar_embedding)

    def coo(self):
        """
        Returns coordinates of nodes.

        Returns
        -------
        x, y : (N,) arrays
            Coordinates of nodes in graph.
        """
        return self.x, self.y

    def node_count(self):
        """
        Returns number of nodes in graph (abbreviated by N).
        """
        return self.N

    def edge_count(self):
        """
        Returns number of edges in graph (abbreviated by E).
        """
        return self.E

    def get_l_cycles(self, to_list=True):
        """
        Returns for all l-cycles the nodes it traverses in order.

        Parameters
        ----------
        to_list=False :
            If true, output is in form of list-of-lists, otherwise concatenated array.

        Returns
        -------
        nodes :  list-of-lists or array
             For all l-cycles the nodes it traverses in order.
        lengths : (FR,) int array
            Length of every l-cycle.
        """
        self._assign_faces()
        self._assign_face_nodes()
        nodes, lengths = self.face_nodes, self.face_lengths
        if to_list:
            return VarRowArray(lengths).to_list(nodes), lengths
        return nodes, lengths

    def get_face_cycles(self, to_list=True):
        """
        Returns for all face-cycles the nodes it traverses in order. If the graph is planar,
        these enclose individual faces.

        Parameters
        ----------
        to_list=False :
            If true, output is in form of list-of-lists, otherwise concatenated array.

        Returns
        -------
        nodes :  list-of-lists or array
             For all face-cycles the nodes it traverses in order.
        lengths : (F,) int array
            Length of every face-cycle.
        """
        self._assign_faces()
        self._assign_face_nodes()
        self._assign_boundary_faces()
        nodes = self.faces_v_array.delete_rows(self.face_nodes, self.boundary_face_indices)
        lengths = np.delete(self.face_lengths, self.boundary_face_indices)
        if to_list:
            return VarRowArray(lengths).to_list(nodes), lengths
        return nodes, lengths

    def face_count(self):
        """
        Returns number of face cycles in graph (abbreviated by F). In a planar
        graph this equals the number of faces.
        """
        self._assign_faces()
        self._assign_boundary_faces()
        return self.F - len(self.get_boundary_faces())

    def l_cycle_count(self):
        """
        Returns number of l-cycles in graph. (abbreviated by FR)
        """
        self._assign_faces()
        return self.F

    def is_planar_embedding(self):
        """
        Returns if graph is planar embedding, which is true if edges only intersect at their endpoints.
        """
        self._assign_faces()
        return self.node_count() + self.face_count() == self.edge_count() + 1

    def get_face_areas(self):
        """
        Returns areas of face-cycles. These correspond to face-areas is the graph
        is planar.

        Returns
        -------
        areas :  (F,) array
            Areas of face-cycles.
        """
        self._assign_areas()
        self._assign_boundary_faces()
        return np.delete(self.areas, self.boundary_face_indices)

    def get_l_cycle_areas(self):
        """
        Returns signed areas of l-cycles.

        Returns
        -------
        areas :  (FR,) array
            Signed areas of l-cycles.
        """
        self._assign_areas()
        return self.areas

    def get_l_cycle_centroids(self):
        # TODO
        """
        Returns centroids of l_cycles in graph

        Returns
        -------
        x, y :  (F,) arrays
            Returns coordinates of centroids of l_cycles.
        """
        six_times_area = 6.0 * self.get_l_cycle_areas()
        mask = np.isclose(six_times_area, 0.0)
        six_times_area[mask] = 2.0 * self.face_lengths[mask]
        long_mask = self.faces_v_array.at_out_index(mask)
        X = self.x[self.face_nodes] + self.x[self.roll_ids]
        Y = self.y[self.face_nodes] + self.y[self.roll_ids]
        centroid_x = self.faces_v_array.sum(np.where(long_mask, X, self.determinant * X)) / six_times_area
        centroid_y = self.faces_v_array.sum(np.where(long_mask, Y, self.determinant * Y)) / six_times_area
        return centroid_x, centroid_y

    def get_face_centroids(self):
        """
        Returns centroids of faces in graph

        Returns
        -------
        x, y :  (F,) arrays
            Returns coordinates of centroids of faces.
        """
        centroid_x, centroid_y = self.get_l_cycle_centroids()
        self._assign_boundary_faces()
        return np.delete(centroid_x, self.boundary_face_indices), \
               np.delete(centroid_y, self.boundary_face_indices)

    def get_num_components(self):
        """
        Returns number of connected components of graph.
        """
        return scipy.sparse.csgraph.connected_components(
            self.adjacency_matrix(), directed=False, return_labels=False)

    def get_components(self):
        """
        Returns components of graph.

        Returns
        -------
        components :  (N,) int array in range(num_components)
            Components of graph.
        """
        _, components = scipy.sparse.csgraph.connected_components(
            self.adjacency_matrix(), directed=False, return_labels=True)
        return components

    def split_components(self):
        """
        Returns list of individual components of graph as separate EmbeddedGraph objects.
        """
        components = self.get_components()
        num_components = np.max(components) + 1
        return [self.remove_nodes(np.flatnonzero(components != c)) for c in range(num_components)]

    def get_boundary_faces(self):
        """
        Returns indices of l_cycles that are boundary faces.
        """
        self._assign_boundary_faces()
        return self.boundary_face_indices

    def l_cycle_matrix(self):
        """
        Returns cycle matrix build with l_cycles.

        Cycle matrix is an (F, E) sparse matrix where the i-th row corresponds to the
        i-th l_cycle. The entry is +1 or -1 if the cycle contains that edge.

        The value +1 is assigned if traversing the cycle counter clockwise first encounters
        the endpoint node of the edge with lowest id. If not, -1 is assigned.
        """
        self._assign_faces()
        return self._cycle_matrix(self.face_edges, self.face_lengths, self.edge_permute, self.edge_flip)

    def _cycle_matrix(self, face_edges, face_lengths, permute, edge_flip):
        E, F = self.edge_count(), len(face_lengths)
        indptr = np.append([0], np.cumsum(face_lengths))
        indices, data = permute[face_edges % E], 1 - 2 * (face_edges // E)
        data = np.where(edge_flip[indices], -data, data)
        return scipy.sparse.csr_matrix((data, indices, indptr), shape=(F, E)).tocsc()

    def face_cycle_matrix(self):
        """
        Returns cycle matrix build with face cycles

        Cycle matrix is an (F, E) sparse matrix where the i-th row corresponds to the
        i-th face. The entry is +1 or -1 if the cycle contains that edge.

        The value +1 is assigned if traversing the cycle counter clockwise first encounters
        the endpoint node of the edge with lowest id. If not, -1 is assigned.
        """
        self._assign_faces()
        self._assign_boundary_faces()
        nb_face_edges = self.face_edges[self.faces_v_array.get_item(rows=self._non_boundary_mask())]
        nb_face_lengths = self.face_lengths[self._non_boundary_mask()]
        return self._cycle_matrix(nb_face_edges, nb_face_lengths, self.edge_permute, self.edge_flip)

    # def cut_space_matrix_old(self):
    #     """
    #     Returns cut-space matrix.
    #
    #     Cut-space matrix is an (N, E) sparse matrix where the i-th row corresponds to the
    #     i-th node. The entry is +1 or -1 if the node is an endpoint of that edge. It is +1
    #     if the node is the endpoint with highest idx, -1 otherwise.
    #     """
    #     E, N = self.edge_count(), self.node_count()
    #     row = np.concatenate((self.node1, self.node2))
    #     col = self.edge_permute[np.concatenate((np.arange(E), np.arange(E)))]
    #     data = np.concatenate((-np.ones(E), np.ones(E)))
    #     data = np.where(self.edge_flip[col], -data, data)
    #     return scipy.sparse.coo_matrix((data, (row, col)), shape=(N, E)).tocsc()

    def cut_space_matrix(self):
        """
        Returns cut-space matrix.

        Cut-space matrix is an (N, E) sparse matrix where the i-th row corresponds to the
        i-th node. The entry is +1 or -1 if the node is an endpoint of that edge. It is +1
        if the node is the endpoint with highest idx, -1 otherwise.
        """
        E, N = self.edge_count(), self.node_count()
        n1, n2 = self.get_edges()
        row = np.concatenate((n1, n2))
        col = np.concatenate((np.arange(E), np.arange(E)))
        data = np.concatenate((-np.ones(E), np.ones(E)))
        return scipy.sparse.coo_matrix((data, (row, col)), shape=(N, E)).tocsc()


    def adjacency_matrix(self):
        """
        Returns adjacency matrix.

        Adjacency matrix is an (N, N) sparse symmetric matrix where a node-pair
        entry is 1 if the nodes share an edge, 0 if not.
        """
        E, N = self.edge_count(), self.node_count()
        data = np.ones(2 * E)
        row = np.append(self.node1, self.node2)
        col = np.append(self.node2, self.node1)
        return scipy.sparse.coo_matrix((data, (row, col)), shape=(N, N)).tocsc()

    def get_common_edge_of_l_cycles(self, cycle1, cycle2, return_orientation=False):
        """
        Returns index of an edge occurring in both cycles if it exists; otherwise -1.

        Parameters
        ----------
        cycle1, cycle2 : int arrays in range(F)
            Indices referring to l_cycles.
        return_orientation=False :
            If True, orientation of edge with respect to cycle pair is returned.

        Returns
        -------
        shared_edge : array in range(E) with same size as cycle1,2
            Edge shared by l_cycle1 and l_cycle2. If multiple exist;
            returns one with the lowest index.
        orientation : bool array with same size as cycle1,2 (optional)
            True if cycle1 passes edge counterclockwise first encountering its
            node with lowest index.
        """
        self._assign_faces()
        f = self.faces_v_array.rows()[np.argsort(self.face_edges)].astype(np.int64)
        E = self.edge_count()
        f1, f2 = f[:E], f[E:2 * E]
        mask = f1 < f2
        f1, f2 = np.where(mask, f1, f2),  np.where(mask, f2, f1)

        sorter = np.lexsort((f2, f1))
        f1, f2 = f1[sorter],  f2[sorter]
        A = np.array(list(zip(f1, f2)), dtype=[('f1', 'int'), ('f2', 'int')])

        cycle1, cycle2 = np.array(cycle1, dtype=int), np.array(cycle2, dtype=int)
        input_shape = cycle1.shape
        cycle1, cycle2 = cycle1.ravel(), cycle2.ravel()
        mask2 = cycle1 < cycle2
        fmin, fmax = np.where(mask2, cycle1, cycle2).astype(np.int64), np.where(mask2, cycle2, cycle1).astype(np.int64)
        V = np.array(list(zip(fmin, fmax)), dtype=[('fmin', 'int'), ('fmax', 'int')])

        edge_ids = np.searchsorted(A, V)
        edge_ids[edge_ids >= E] = E - 1
        if return_orientation:
            orientation = mask[sorter][edge_ids] ^ ~mask2

        found_mask = (f1[edge_ids] != fmin) | (f2[edge_ids] != fmax)
        edge_ids = sorter[edge_ids]

        edge_ids[found_mask] = -1

        ids = self.edge_permute[edge_ids.reshape(input_shape)]
        if return_orientation:
            orientation[found_mask] = False
            return ids, orientation.reshape(input_shape) ^ self.edge_flip[ids]
        else:
            return ids

    def permute_nodes(self, permutation):
        """
        Permute node order. Because edges refer to nodes by their position,
        this also changes node1 and node2.
        """
        permutation = np.array(permutation, dtype=int)
        if not np.all(np.sort(permutation) == np.arange(self.node_count())):
            raise ValueError("invalid permutation")
        self.x = self.x[permutation]
        self.y = self.y[permutation]
        inv_perm = np.argsort(permutation)
        self.node1 = inv_perm[self.node1]
        self.node2 = inv_perm[self.node2]
        self.node1, self.node2 = self._de_sort_edges()
        self._sort_edges()
        self._recompute_faces()

    def permute_faces(self, permutation):
        """
        Permute face order.
        """
        permutation = np.array(permutation, dtype=int)
        if not np.all(np.sort(permutation) == np.arange(self.l_cycle_count())):
            raise ValueError("invalid permutation")
        self.face_permutation = self.face_permutation[permutation]
        self._recompute_faces()

    def face_dual_graph(self):
        A = self.face_cycle_matrix()
        adj = (A @ A.T).tocoo()
        mask = adj.row < adj.col
        return EmbeddedGraph(*self.get_face_centroids(), adj.row[mask], adj.col[mask])

    def locate_faces(self, x, y):
        """
        Get faces whose centroids are closest to queried coordinate.
        Graph must be planar embedding.

        Attributes
        ----------
        x, y: arrays :
            Coordinates at which one wants to locate faces.

        Returns
        -------
        face_ids : int array with same size as x in range(F)
            Returns indices of located faces.
        """
        if not self.is_planar_embedding():
            raise ValueError("Only works for planar embedding")
        locator = scipy.spatial.KDTree(np.stack(self.get_face_centroids(), axis=-1))
        _, face_ids = locator.query(np.stack(np.broadcast_arrays(x, y), axis=-1), k=1)
        return face_ids, locator

    def plot(self, fig=None, ax=None, show_cycles=True, cycles="face_cycles", figsize=[5, 5],
             show_node_ids=False, show_edge_ids=False, show_face_ids=False,
             face_shrink_factor=0.9, markersize=5, linewidth=1):

        def _face_line(x, y, n, xcn, ycn, f_shrink):
            xp = f_shrink * x[np.append(n, n[0])] + (1 - f_shrink) * xcn
            yp = f_shrink * y[np.append(n, n[0])] + (1 - f_shrink) * ycn
            return np.stack((xp, yp), axis=1)

        if fig is None and ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        if ax is None:
            ax = fig.add_axes()
        if fig is None:
            fig = ax.get_figure()
        self.fig, self.ax = fig, ax

        x, y = self.coo()
        n1, n2 = self.get_edges()
        lines = [((x[n1[i]], y[n1[i]]), (x[n2[i]], y[n2[i]])) for i in range(len(n1))]
        lc = LineCollection(lines, colors=[0.5,0.5,0.5], linewidths=linewidth)
        self.ax.add_collection(lc)
        self.ax.plot(x, y, color=[0,0,0], marker="o", markerfacecolor=[0,0,0], linestyle="None", markersize=markersize)

        if show_node_ids:
            for i, (xn, yn) in enumerate(zip(x, y)):
                self.ax.text(xn, yn, i.__str__())
        if show_edge_ids:
            x1, y1,  x2, y2 = x[n1], y[n1], x[n2], y[n2]
            for i, (xn, yn) in enumerate(zip(0.5 * (x1 + x2), 0.5 * (y1 + y2))):
                self.ax.annotate(i.__str__(), (xn, yn), color=[0.3, 0.5, 0.9], ha='center', va='center')
        if show_cycles:
            lines = []
            if cycles == "face_cycles":
                xc, yc = self.get_face_centroids()
                pn, _ = self.get_face_cycles(to_list=True)
                for i, (xcn, ycn, n) in enumerate(zip(xc, yc, pn)):
                    lines += [_face_line(x, y, n, xcn, ycn, face_shrink_factor)]
                    if show_face_ids:
                        self.ax.annotate(i.__str__(), (xcn, ycn), color=[1, 0.5, 0.2], ha='center', va='center')
            if cycles == "l_cycles":
                xc, yc = self.get_l_cycle_centroids()
                pn, _ = self.get_l_cycles(to_list=True)
                b_mask = ~self._non_boundary_mask()
                lines = []
                for i, (xcn, ycn, n) in enumerate(zip(xc, yc, pn)):
                    if b_mask[i]:
                        xp, yp = x[n], y[n]
                        self.ax.plot(np.append(xp, xp[0]), np.append(yp, yp[0]),
                                     color=[0.2, 0.5, 1], linewidth=linewidth)
                        if show_face_ids:
                            self.ax.annotate(i.__str__(), (xcn, ycn), color=[0.2, 0.5, 1], ha='center', va='center')
                    else:
                        lines += [_face_line(x, y, n, xcn, ycn, face_shrink_factor)]
                        if show_face_ids:
                            self.ax.annotate(i.__str__(), (xcn, ycn), color=[1, 0.5, 0.2], ha='center', va='center')

            lc1 = LineCollection(lines, colors=[1, 0.5, 0.2], linewidths=linewidth)
            self.ax.add_collection(lc1)
        return self.fig, self.ax

    def shortest_path(self, node1, node2, to_list=True):
        """
        Returns nodes encountered along a shortest path from node1 to node2 through graph.

        - paths include node1 and node2.
        - If node1, node2 are (P,) arrays, returns P paths where path i goes from node1[i] to node2[i].

        Parameters
        ----------
        node1, node2 : (P,) array
            Start- and end-nodes of paths.
        to_list=False :
            If true, output is in form of list-of-lists, otherwise concatenated array.

        Returns
        -------
        paths :  list-of-lists or array
             All nodes that are traversed in path
        lengths : (FR,) int array
            Length of every l-cycle.
        """
        node1, node2 = np.atleast_1d(node1), np.atleast_1d(node2)
        P = len(node2)
        up1, idx = np.unique(node2, return_inverse=True)
        _, pred = scipy.sparse.csgraph.shortest_path(self.adjacency_matrix(), return_predecessors=True, indices=up1)
        K = pred.shape[1]
        pred[pred < 0] = K
        pred = np.concatenate((pred, K * np.ones((pred.shape[0], 1), dtype=int)), axis=1)
        p = np.array(node1).copy()
        out = np.zeros((5, P), dtype=int)
        out[0, :] = node1
        k = 1
        while not np.all(p == K):
            if k >= (out.shape[0] - 1):
                out = np.concatenate((out, 0 * out), axis=0)
            p = pred[(idx, p)]
            out[k, :] = p
            k += 1
        out = out[:k, :]
        path_lengths = np.argmax(out == K, axis=0)
        out = out.T
        paths = out[out < K].ravel()
        if to_list:
            return VarRowArray(path_lengths).to_list(paths), path_lengths
        return paths, path_lengths

    def _assert_edges_correct_shape(self):
        if len(self.node1) != len(self.node2):
            raise ValueError("node1 and node2 must be same size")

    def _assert_edges_contain_existing_nodes(self):
        if np.any((self.node1 < 0) | (self.node1 >= self.N) | (self.node2 < 0) | (self.node2 >= self.N)):
            raise NodeNotExistError("node1,2 values must be in range(N)")

    def _assert_single_component(self):
        if self.get_num_components() != 1:
            raise NotSingleComponentError()

    def _assert_planar_embedding(self):
        if not self.is_planar_embedding():
            raise NotPlanarEmbeddingError()

    def _assert_no_self_loop(self):
        if np.any(self.node1 == self.node2):
            raise SelfLoopError("no edge is allowed to have identical end nodes.")

    def _assert_nonsimple(self):
        if np.any((self.node1[:-1] == self.node1[1:]) & (self.node2[:-1] == self.node2[1:])):
            raise NonSimpleError("no duplicate edges allowed.")

    def _assign_edge_v_array(self):
        self.edge_v_array = VarRowArray(np.diff(np.append(np.flatnonzero(np.roll(self.node1, 1) - self.node1), len(self.node1))))
        return self.edge_v_array

    def _sort_edges(self):
        self.edge_flip = self.node1 > self.node2
        nn1 = np.where(self.edge_flip, self.node2, self.node1)
        nn2 = np.where(self.edge_flip ,self.node1, self.node2)
        self.edge_permute = np.lexsort((nn2, nn1))
        self.edge_permute_inverse = np.argsort(self.edge_permute)
        self.node1, self.node2 = nn1[self.edge_permute], nn2[self.edge_permute]

    def _de_sort_edges(self):
        nn1, nn2 = self.node1[self.edge_permute_inverse], self.node2[self.edge_permute_inverse]
        return np.where(self.edge_flip, nn2, nn1), np.where(self.edge_flip, nn1, nn2)

    def _reset_precomputed_quantities(self):
        self.F = None
        self.faces_v_array = None
        self.face_edges = None
        self.face_nodes = None
        self.face_lengths = None
        self.determinant = None
        self.areas = None
        self.roll_ids = None
        self.boundary_face_indices = None

    def _assign_faces(self):
        if self.face_edges is None:
            self.face_edges, self.face_lengths = self._construct_l_cycles()
            self.faces_v_array = VarRowArray(self.face_lengths)
            self.F = len(self.face_lengths)
        if self.face_permutation is None:
            self.face_permutation = np.arange(self.F)

    def _assign_face_nodes(self):
        self._assign_faces()
        if self.face_nodes is None:
            e_idx = self.face_edges % self.edge_count()
            self.face_nodes = np.where(self.face_edges < self.edge_count(), self.node1[e_idx], self.node2[e_idx])

    def _assign_roll_ids(self):
        self._assign_face_nodes()
        if self.roll_ids is None:
            self.roll_ids = self.face_nodes[self.faces_v_array.roll(1)]

    def _assign_determinant(self):
        self._assign_roll_ids()
        if self.determinant is None:
            self.determinant = self.x[self.roll_ids] * self.y[self.face_nodes] - self.x[self.face_nodes] * self.y[self.roll_ids]

    def _assign_areas(self):
        self._assign_determinant()
        if self.areas is None:
            self.areas = 0.5 * self.faces_v_array.sum(self.determinant)

    def _assign_boundary_faces(self):
        self._assign_areas()
        if self.boundary_face_indices is None:
            self.boundary_face_indices = np.flatnonzero((self.areas < 0) | np.isclose(self.areas, 0.0))

    def _recompute_faces(self):
        self._reset_precomputed_quantities()
        permutation = self.face_permutation
        inv_permutation = np.argsort(permutation)
        self._assign_faces()
        self._assign_face_nodes()
        self._assign_areas()
        self._assign_determinant()
        self._assign_roll_ids()
        self._assign_boundary_faces()
        self.face_lengths = self.face_lengths[permutation]
        self.face_edges = self.faces_v_array.permute_rows(permutation, self.face_edges)
        self.areas = self.areas[permutation]
        self.determinant = self.faces_v_array.permute_rows(permutation, self.determinant)
        self.roll_ids = self.faces_v_array.permute_rows(permutation, self.roll_ids)
        self.boundary_face_indices = inv_permutation[self.boundary_face_indices]
        self.face_nodes = self.faces_v_array.permute_rows(permutation, self.face_nodes)
        self.faces_v_array = VarRowArray(self.faces_v_array.counts[permutation])

    def _get_edge_map(self):
        """
        Computes a one-to-one map over directed edges, where the image contains the edge
        which is encountered next when traversing the graph moving counter-clockwise.

        Used to generate faces by repeating indexing: e_(i+1) = map[e_i]

        map: np.arange(2*E) -> sorted_out_edge_directed

        Output:
        sorted_out_edge_directed    (2*E,) in range(2*E)
        """

        counter_clockwise = True
        edge_count = self.edge_count()
        ns = np.append(self.node1, self.node2)

        # construct neighbour structure
        nodes, count = np.unique(ns, return_counts=True)
        neighbours = VarRowArray(count)
        neighbour_edges = np.tile(np.arange(edge_count), 2)[np.argsort(ns)]
        neighbour_node_self = nodes[neighbours.rows()]
        neighbour_node_other = np.where(self.node1[neighbour_edges] == neighbour_node_self,
                                        self.node2[neighbour_edges], self.node1[neighbour_edges])

        # sort neighbour in-dimension by ascending angle
        angles = np.arctan2(self.y[neighbour_node_other] - self.y[neighbour_node_self],
                            self.x[neighbour_node_other] - self.x[neighbour_node_self])
        angle_arg_sort = np.argsort(3 * np.pi * neighbour_node_self.astype(np.double) + angles)
        neighbour_edges = neighbour_edges[angle_arg_sort]
        neighbour_node_other = neighbour_node_other[angle_arg_sort]

        def to_directed(edge_nr, edge_direction, edge_count):
            return edge_nr + edge_count * (1 - edge_direction.astype(int))

        # create combined-index for input edges of the map
        neighbour_edges_direction = neighbour_node_other < neighbour_node_self
        in_edge_combined_index = to_directed(neighbour_edges, neighbour_edges_direction, edge_count)

        # find the map from every (combined index) edge to the next edge in (counter-)-clockwise direction
        edges_map = neighbours.roll() if counter_clockwise else neighbours.roll(-1)

        # find the combined index of the output edges of the map
        out_edge_directed = to_directed(neighbour_edges, ~neighbour_edges_direction, edge_count)[edges_map]

        # sort the edge-map based on the (combined index of the) input edges
        sorted_out_edge_directed = out_edge_directed[np.argsort(in_edge_combined_index)]

        return sorted_out_edge_directed

    def _construct_l_cycles(self):
        """
        Computes l-cycles of the embedded graph. This is done using repeated iteration over the
        one-to-one map over directed edges computed with _get_edge_map(). It starts with all
        edges pointing from lowest to highest node to ensure all cycles are found.

        Used to generate cycles by repeating indexing: e_(i+1) = map[e_i]

        map: np.arange(2*E) -> sorted_out_edge_directed

        Output:
        sorted_out_edge_directed    (2*E,) in range(2*E)
        """

        map = self._get_edge_map()
        edge_ids = np.arange(self.edge_count())
        cycles = -np.ones((3, self.edge_count()), dtype=int)
        cycles[0, :] = edge_ids
        cycle_lengths = np.ones(self.edge_count(), dtype=int)
        current_cycle_length, out_cycles, out_cycle_lengths = 1, np.zeros(0, dtype=int), np.zeros(0, dtype=int)

        # iteration doing counter-clockwise walks through the graphs starting from each junction
        while len(edge_ids) > 0:
            edge_ids = map[edge_ids]
            is_terminated = edge_ids == cycles[0, :]
            if np.any(is_terminated):
                out_cycle_lengths, out_cycles = self._store_terminated_cycles(cycles[:current_cycle_length, is_terminated].T,
                                                                            out_cycle_lengths, out_cycles)
                cycles, cycle_lengths = cycles[:, ~is_terminated], cycle_lengths[~is_terminated]
                edge_ids = edge_ids[~is_terminated]
            cycles[current_cycle_length, :] = edge_ids
            current_cycle_length += 1
            if cycles.shape[0] == (current_cycle_length):
                cycles = np.append(cycles, -np.ones(cycles.shape, dtype=int), axis=0)
        return out_cycles, out_cycle_lengths

    def _store_terminated_cycles(self, terminated_cycles, out_cycle_lengths, out_cycles):
        # roll cycles until the lowest node index is in the first column (so its easier to remove duplicate cycles)
        cycle_cnt, cycle_len = terminated_cycles.shape
        cycles = terminated_cycles[np.arange(cycle_cnt)[:, None], (
                    np.arange(cycle_len) + np.argmin(terminated_cycles, axis=1)[:, None]) % cycle_len]

        _, idx = np.unique(cycles[:, 0], return_index=True)
        cycles = cycles[idx, :]
        #
        # cycles = np.unique(cycles, axis=0)

        out_cycle_lengths = np.append(out_cycle_lengths, cycle_len * np.ones(cycles.shape[0], dtype=int))
        out_cycles = np.append(out_cycles, cycles.flatten())
        return out_cycle_lengths, out_cycles

    def _non_boundary_mask(self):
        self._assign_boundary_faces()
        out = np.ones(self.l_cycle_count(), dtype=bool)
        out[self.boundary_face_indices] = False
        return out

    def _cycle_space_solve_for_integral_x(self, b):
        """
        Solves the equation: A @ x = b (where A = cycle_matrix, without boundary faces).
        If b is integral (contain only integers), the output array x will also be integral.

        input:  b (..., F)
        output: x (..., E)

        Notes:
            - The equation is underdetermined, so the solution x is not unique.
        """
        self._assign_boundary_faces()
        if (self.get_num_components() != 1) or (not self.is_planar_embedding()):
            raise ValueError("only implemented for single component planar embedding")

        E, F = self.edge_count(), self.face_count()
        b = np.array(b, dtype=np.double)
        b_shape = list(b.shape)
        b = b.reshape(-1, F)
        b_tally = b.copy()

        # insert boundary face in b_tally
        b_idx = self.boundary_face_indices[0]
        b_tally = np.concatenate((b_tally[:, :b_idx], np.zeros((b_tally.shape[0], 1)), b_tally[:, b_idx:]), axis=1)
        # do depth first search, resulting in cur (current node of tree) and prev (parent node of cur)
        A = self.l_cycle_matrix()
        cur, predecessor = scipy.sparse.csgraph.depth_first_order(A @ A.T, b_idx)
        prev = predecessor[cur]
        prev[0] = -1

        # find map of edge between pair of faces. signs returns +1 if face1 counterclockwise passes resulting edge in its own direction.
        juncs, orientation_in_face_1 = self.get_common_edge_of_l_cycles(cur, prev, return_orientation=True)
        sgns = (-1 + 2 * orientation_in_face_1.astype(np.double))

        # construct x at each edge by passing through the tree in reverse.
        x = np.zeros((b.shape[0], E), dtype=b.dtype)
        for i in reversed(range(F+1)):
            if prev[i] >= 0:
                b_tally[:, prev[i]] += b_tally[:, cur[i]]
                x[:, juncs[i]] += b_tally[:, cur[i]] * sgns[i]
                b_tally[:, cur[i]] = 0

        # check if resulting x solves A @ x == b
        b_shape[-1] = E
        mask = self._non_boundary_mask()
        if not np.allclose((A @ x.T)[mask, :], b.T):
            raise ValueError("failed integral solve of cycle space linear problem")

        # return x
        return x.reshape(tuple(b_shape))


class EmbeddedSquareGraph(EmbeddedGraph):

    """
    Generate Embedded graph for a square lattice.

    Parameters
    ----------
    count_x, count_y : int scalar
        Node count of square lattice in x- and y-direction respectively.
    x_scale, y_scale=1.0 : float scalar
        Stretch factors for square lattice in x- and y-direction respectively.
        Bottom-left node is at origin.
    """
    def __init__(self, count_x, count_y,  x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        n1 = np.concatenate((idx[:, 0:-1].ravel(), idx[0:-1, :].ravel()))
        n2 = np.concatenate((idx[:, 1:].ravel(), idx[1:, :].ravel()))
        super().__init__(x * x_scale, y * y_scale, n1, n2)


class EmbeddedHoneycombGraph(EmbeddedGraph):
    """
    Generate Embedded graph for a honeycomb lattice.

    Parameters
    ----------
    count_x, count_y : int scalar
        The amount of times the unit cell (with shape /¯\ and has 4 nodes) is
        repeated in the x- and y-direction; generating a honeycomb lattice.
    x_scale, y_scale=1.0 : float scalar
        Stretch factors for honeycomb lattice in x- and y-direction respectively.
        Bottom-left node is at origin.
    """
    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        x1, y1 = 3.0 * x, np.sqrt(3.0) * y
        nodes_x = np.concatenate((x1, x1 + 0.5, x1 + 1.5, x1 + 2), axis=0).ravel()
        nodes_y = np.concatenate((y1, y1 + np.sqrt(0.75), y1 + np.sqrt(0.75), y1), axis=0).ravel()
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        s = count_x * count_y
        nodes1 = (idx,   idx[:-1, :]+s, idx+s,   idx[:-1, :]+2*s, idx+2*s, idx[:, :-1]+3*s)
        nodes2 = (idx+s, idx[1:, :],    idx+2*s, idx[1:, :]+3*s,  idx+3*s, idx[:, 1:])
        nodes1 = np.concatenate(tuple([n1.flatten() for n1 in nodes1])).ravel()
        nodes2 = np.concatenate(tuple([n2.flatten() for n2 in nodes2])).ravel()
        # remove_node_ids = [0, idx[0, -1] + 3 * s]
        # self.remove_nodes(remove_node_ids)
        super().__init__(nodes_x * x_scale, nodes_y * y_scale, nodes1, nodes2)

class EmbeddedTriangularGraph(EmbeddedGraph):
    """
    Generate Embedded graph for a triangular lattice.

    Parameters
    ----------
    count_x, count_y : int scalar
        The amount of times the unit cell (with shape / and has 2 nodes) is
        repeated in the x- and y-direction; generating a triangular lattice.
    x_scale, y_scale=1.0 : float scalar
        Stretch factors for triangular lattice in x- and y-direction respectively.
        Bottom-left node is at origin.
    """
    def __init__(self, count_x, count_y, x_scale=1.0, y_scale=1.0):
        y, x = np.mgrid[0:count_y, 0:count_x]
        x1, y1 = x, np.sqrt(3.0) * y
        nodes_x = np.concatenate((x1, x1 + 0.5), axis=0)
        nodes_y = np.concatenate((y1, y1 + np.sqrt(0.75)), axis=0)
        idx = np.arange(count_x * count_y).reshape(count_y, count_x)
        s = count_x * count_y
        nodes1 = (idx,   idx[:, :-1], idx[:-1, :]+s, idx[:-1, :-1]+s, idx[:, :-1]+s, idx[:, 1:])
        nodes2 = (idx+s, idx[:, 1:],  idx[1:, :],    idx[1:, 1:],     idx[:, 1:]+s,  idx[:, :-1]+s)
        nodes1 = np.concatenate(tuple([n1.flatten() for n1 in nodes1]))
        nodes2 = np.concatenate(tuple([n2.flatten() for n2 in nodes2]))
        super().__init__(nodes_x * x_scale, nodes_y * y_scale, nodes1, nodes2)




import time

from pyjjasim import *

import matplotlib.pyplot as plt

"""
EXAMPLE 4: Parameter optimization

Compute maximal frustration bounds and maximal current 
"""

if __name__ == "__main__":
    # define arrays
    array = SquareArray(12, 12)
    # Ih, Iv = 1.0 * array.horizontal_junctions(), 1.0 * array.vertical_junctions()

    # compute frustration bounds for zero vortex state
    prob = StaticProblem(array, frustration=1, current_sources=0)
    (smallest_f, largest_f), (s_config, l_config), _ = prob.compute_frustration_bounds()
    print(f"smallest and largest frustration for which the zero-vortex state exists: {smallest_f}, {largest_f}")
    s_config.plot(title=f"minimal frustration in zero vortex state (f={np.round(smallest_f, 4)})")

    # compute maximal current
    prob = StaticProblem(array, frustration=0, current_sources=array.current_base(angle=0))
    I_factor, max_I_config, info = prob.compute_maximal_current()
    np.set_printoptions(linewidth=10000000)
    print(f"largest current factor {I_factor} at which the zero-vortex state exists at zero frustration")
    max_I_config.plot(title=f"maximal current in zero vortex state (I_factor={I_factor:.3f})")

    # compute extermum in Is-f space using compute_stable_region()
    prob = StaticProblem(array, frustration=1, current_sources=array.current_base(angle=np.pi/2))
    f_factors, I_factors, _, _ = prob.compute_stable_region()
    plt.subplots()
    plt.plot(f_factors, I_factors)
    plt.xlabel("frustration factors")
    plt.ylabel("current source factors")
    plt.title("Region in f-I parameter space where the zero-vortex state is stable")

    # compute direction dependent maximal current using compute_maximal_parameter()
    prob = StaticProblem(array)
    angles = np.linspace(0, 2*np.pi, 33)
    I_factor = np.zeros(len(angles))
    tic = time.perf_counter()
    for i, angle in enumerate(angles):
        prob_func = lambda x: prob.new_problem(current_sources=x * array.current_base(angle=angle))
        I_factor[i], _, _, _ = compute_maximal_parameter(prob_func)
    print(time.perf_counter() - tic)
    plt.subplots()
    plt.plot(np.cos(angles) * I_factor, np.sin(angles) * I_factor, marker="o")
    plt.xlabel("horizontal current")
    plt.ylabel("vertical current")
    plt.title("angle dependent maximal current for zero-vortex state")
    plt.show()



from multiprocessing import Pool

from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")


"""
EXAMPLE 11: Multithreading

"""

def func(L):
    N = 10
    sq_array = SquareArray(N, N)
    sq_array.set_inductance_factors(L)
    prob = StaticProblem(sq_array, current_sources=sq_array.current_base(angle=0))
    f, I, _, _ = prob.compute_stable_region(angles=np.linspace(0, 2*np.pi, 61))
    return f, I

if __name__ == '__main__':
    L = [0, 0.5, 1, 2]
    with Pool() as pool:
        out = pool.map(func, np.arange(4))

    for i in range(4):
        plt.plot(out[i][0], out[i][1], label=f"beta_L={L[i]}")
    plt.xlabel("frustration")
    plt.ylabel("maximal array current")
    plt.legend()
    plt.show()




from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

"""
Dynamic example 3: Moving a vortex with current pulse. Used in rapid-single-flux-quanum (RSFQ) technology.
"""

from scipy.stats import norm

if __name__ == "__main__":
    # define array
    N = 20
    sq_array = SquareArray(N, N)

    # define problem
    dt = 0.05
    Nt = 2000
    ts = np.arange(0, Nt, 3)
    Ih = sq_array.current_base(angle=0)
    amp = 2
    Is = amp * Ih[:, None, None] * norm(loc=30, scale=1.0).pdf(np.arange(Nt) * dt)

    # define initial condition with vortex in centre
    n = np.zeros(sq_array.face_count(), dtype=int)
    n[sq_array.locate_faces((N - 1) / 2, (N - 1) / 2)] = 1
    init, _, _ = StaticProblem(sq_array, vortex_configuration=n).compute()
    init_th = init.get_theta()[:, None]

    problem = TimeEvolutionProblem(sq_array, time_step=dt, time_step_count=Nt, current_sources=Is,
                                   store_time_steps=ts, config_at_minus_1=init_th)

    # do time simulation
    out = problem.compute()

    # animate result
    handles = out.animate(figsize=(8, 8), title="Move vortex with current pulse ")
    plt.show()




from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")

"""
Static Example 7: Randomized square array

First a square array with varying critical current is investigated.

Secondly, a square array with random holes is investigated.

"""

if __name__ == "__main__":

    # EXAMPLE A: RANDOM CRITICAL CURRENT

    np.random.seed(1)

    # define arrays
    N = 12
    array = SquareArray(N, N)

    Ic = 1 + 0.3 * np.random.randn(array.junction_count())
    Ic[Ic < 0.1] = 0.1
    plt.hist(Ic)
    plt.title("histogram of junction Ic")
    plt.xlabel("critical current")
    plt.ylabel("count")

    array.set_critical_current_factors(Ic)

    # define problem parameters
    problem_no_scr = StaticProblem(array, current_sources=array.current_base(angle=0))
    I_factor, net_I, config, _ = problem_no_scr.compute_maximal_current()

    print(f"max net current: {net_I}. For uniform Ic, it would be 12.")
    config.plot(title="maximum current with random Ic")


    # EXAMPLE B: Arrays with holes
    remove_node_count = 20
    array.set_critical_current_factors(1)
    array = array.remove_nodes(np.random.permutation(array.node_count())[:remove_node_count])

    # create junction-current sources such that uniform current is injected
    x, y = array.get_node_coordinates()
    left_nodes = x==0
    right_nodes = x==(N-1)
    Is_node = np.zeros(array.node_count(), dtype=float)
    Is_node[left_nodes] = np.sum(right_nodes)
    Is_node[right_nodes] = -np.sum(left_nodes)
    # assert sum is zero (current conservation):
    print(f"this sum must be zero: {np.sum(Is_node)}")
    Is = node_to_junction_current(array, Is_node)

    problem_no_scr = StaticProblem(array, current_sources=Is)
    I_factor, net_I, config, _ = problem_no_scr.compute_maximal_current()

    print(f"max net current: {net_I}. Without holes, it would be 12.")
    config.plot(title="maximum current with holes")
    plt.show()




from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

"""
Dynamic example 5: Giant shapiro steps

With a biassed AC current, the DC voltage has plateaus. This effect is called shapiro steps for single junctions,
and giant shapiro steps for arrays as the step height is proportional to the array size. 

Of course after the array voltage is normalized, the steps are not so 
giant, and the step heights will be Ifreq.
"""

if __name__ == "__main__":

    # define array
    N = 20
    sq_array = SquareArray(N, N)

    # define problem
    T = 0.01
    dt = 0.05
    Nt = 10000
    ts = [Nt//3, Nt - 1]

    IAmp = 1
    Ifreq = 0.25
    IDC = np.linspace(0, 2, 101)
    Is = lambda i: sq_array.horizontal_junctions()[:, None] * (IDC + np.sin(Ifreq * i * dt))

    prob = TimeEvolutionProblem(sq_array, time_step=dt, time_step_count=Nt,
                                current_sources=Is, temperature=T, store_time_steps=ts)

    # do time simulation
    out = prob.compute()

    # compute array voltage
    th = out.get_theta()[sq_array.horizontal_junctions(), :, :]
    V = np.mean((th[:, :, 1] - th[:, :, 0]) / (dt * (ts[1] - ts[0])), axis=0)

    # plot array voltage
    plt.plot(IDC, V, label='f=0')
    plt.xlabel("DC current")
    plt.ylabel("mean array voltage")
    plt.title("giant shapiro steps in square array")
    plt.legend()
    plt.show()



import numpy as np
import scipy.sparse


class VarRowArray:
    """
    Array structure representing 2D array with variable row lengths.
    The rows are called the "inner" dimension, the columns the "outer".
    One supplies the length of each row with counts.

    example:
    counts        [2 3 1 2]   represents [[., .], [., ., .], [.], [., .]] array
    columns       [0 1 0 1 2 0 0 1]
    rows          [0 0 1 1 1 2 3 3]
    row_length    [2 2 3 3 3 1 2 2]
    __len__()     8
    get_item([1, 0, 2, 1, 3], [1, 1, 0, 2, 1]) -> [3, 1, 5, 4, 7]
    roll(axis=1)  [1 0 4 2 3 5 7 6]
    roll(axis=0)  [6 7 0 1 2 3 4 5]

    """

    def __init__(self, counts):
        """

        :param counts:
        """
        self.counts = np.array(counts, dtype=int)
        self.cum_counts = np.cumsum(self.counts) - self.counts
        self.total = np.sum(self.counts)

    def __len__(self):
        return self.total

    def row_count(self):
        """
        compute row count

        :return:
        """
        return len(self.counts)

    def columns(self, rows=None):
        # if rows is None; returns column indices for all rows
        if rows is None:
            return np.arange(len(self)) - np.repeat(self.cum_counts, self.counts)
        else:
            c = self.counts[rows]
            return np.arange(np.sum(c)) - np.repeat(np.cumsum(c) - c, c)

    def row_ranges(self):
        return self.cum_counts, self.cum_counts + self.counts

    def sum(self, data, axis=1):
        if axis == 0:
            raise ValueError("sum along axis=0 not implemented")
        if axis == 1:
            s_data = np.append([0], np.cumsum(data))
            return s_data[self.cum_counts + self.counts] - s_data[self.cum_counts]
        raise ValueError("axis must be 0 or 1")

    def rows(self):
        return np.repeat(np.arange(len(self.counts)), self.counts)

    def at_out_index(self, data):
        return np.repeat(data, self.counts)

    def row_length(self):
        return np.repeat(self.counts, self.counts)

    def get_item(self, rows=None, columns=None):
        if rows is not None:
            rows = np.arange(self.row_count())[rows]
        if rows is not None and columns is not None:
            return np.array(columns, dtype=int) + self.cum_counts[rows]
        if columns is None:
            columns = self.columns(rows=rows)
        if rows is None:
            rows = self.rows()
        else:
            rows = np.repeat(np.array(rows), self.counts[rows])

        return columns + self.cum_counts[rows]

    def delete_rows(self, data, rows):
        return np.delete(data, self.get_item(rows=rows))

    def roll(self, n=1, axis=1):
        if axis==0:
            if n >= 0:
                return np.roll(np.arange(self.total), np.sum(self.counts[-n:]))
            else:
                return np.roll(np.arange(self.total), -np.sum(self.counts[:-n]))
        if axis==1:
            return self.get_item(columns=(self.columns() - n) % self.row_length())
        raise ValueError("axis must be 0 or 1")

    def to_csr(self, cols, data, num_cols=None):
        if num_cols is None:
            num_cols = np.max(cols)
        indptr = np.append(self.cum_counts, [self.total])
        return scipy.sparse.csr_matrix((data, cols, indptr), shape=(self.row_count(), num_cols))

    def to_list(self, data):
        return [[data[self.cum_counts[row] + col] for col in range(self.counts[row])] for row in range(self.row_count())]

    def permute_rows(self, permutation, data):
        permutation, data = np.array(permutation, dtype=int), np.array(data)
        return data[self.get_item(rows=permutation)]

    def merge(self, other):
        # returns sorter;
        # out = np.zeros(len(self) + len(other))
        # out[sorter] = np.append(data_self, data_other)
        t = self.counts + other.counts
        cum_t = np.cumsum(t) - t
        i1 = self.columns() + np.repeat(cum_t, self.counts)
        i2 = other.columns() + np.repeat(cum_t + self.counts, other.counts)
        return np.append(i1, i2)



if __name__ == "__main__":

    x = VarRowArray([2, 3, 1, 2])
    print(len(x))
    print(x.row_count())
    print(x.columns())
    print(x.rows())
    print(x.roll())
    print(x.roll(axis=0))
    print(x.get_item([1, 0, 2, 1, 3], [1, 1, 0, 2, 1]))
    print(x.to_list([12, 34, 13, 24, 56, 54, 32, 21]))

    data = [2, 6, 4, 5, 4, 6, 7, 8]
    perm = [2, 1, 3, 0]
    print(x.permute_rows(perm, data))



from pyjjasim import *

import matplotlib
matplotlib.use("TkAgg")
import matplotlib.pyplot as plt


"""
EXAMPLE 5: Lattices

Compute maximal current in a square, honeycomb and triangular lattice.
"""

if __name__ == "__main__":
    # define arrays
    square_array = SquareArray(12, 12)
    honeycomb_array = HoneycombArray(8, 12)
    triangular_array = TriangularArray(12, 8)

    # frustration factor
    f = 0.02

    # define physical problems
    prob_sq = StaticProblem(square_array, frustration=f, current_sources=square_array.current_base(angle=0))
    prob_hc = StaticProblem(honeycomb_array, frustration=f, current_sources=honeycomb_array.current_base(angle=0))
    prob_tr = StaticProblem(triangular_array, frustration=f, current_sources=triangular_array.current_base(angle=0))

    square_array.plot()
    honeycomb_array.plot()
    triangular_array.plot()
    # plt.show()

    # compute maximal current
    _, _, config_sq, _ = prob_sq.compute_maximal_current()
    _, _, config_hc, _ = prob_hc.compute_maximal_current()
    _, _, config_tr, _ = prob_tr.compute_maximal_current()

    # plot result
    config_sq.plot()
    config_hc.plot()
    config_tr.plot()
    plt.show()



from pyjjasim import *

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use("TkAgg")

"""
Dynamic example 6: Annealing

Finds low energy states by doing a time evolution with gradually decreasing temperature.

"""

if __name__ == "__main__":

    # define array
    N = 20
    sq_array = SquareArray(N, N)

    # define problem
    f = 0.1
    time_step = 2.0
    interval_steps = 20
    interval_count = 400
    vortex_mobility = 0.001
    problem_count = 50
    start_T = 1.0
    T_factor = 1.03
    problem = AnnealingProblem(sq_array, time_step=time_step, interval_steps=interval_steps,
                               interval_count=interval_count, vortex_mobility=vortex_mobility,
                               frustration=f, problem_count=problem_count,
                               start_T=start_T, T_factor=T_factor)
    # do time simulation
    out = problem.compute()
    vortex_configurations, energies, status, configurations, T = out

    plt.subplots()
    plt.hist(energies)
    plt.xlabel("mean energy per junction")
    plt.ylabel("histogram count")

    # plot temperature evolution
    plt.subplots()
    plt.plot(time_step * interval_steps * np.arange(interval_count), T)
    plt.xlabel("time")
    plt.ylabel("annealing temperature")

    lowest_state = np.argmin(energies)
    configurations[lowest_state].plot()
    plt.title(f"lowest found state at E={np.round(energies[lowest_state], 5)}")
    plt.show()



import time

import matplotlib
import numpy as np
from scipy.spatial import Voronoi

import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection, LineCollection
from matplotlib.colors import Normalize, LogNorm
import matplotlib.animation as animation
import matplotlib.cm as cm

from pyjjasim import Circuit

from pyjjasim.static_problem import StaticConfiguration
from pyjjasim.time_evolution import TimeEvolutionResult

__all__ = ["CircuitPlot", "ConfigPlot", "CircuitMovie", "TimeEvolutionMovie"]

# TODO:


class CircuitPlot:

    """Plots a circuit configuration.

    Allows one to show node quantities, arrow quantities, face quantities and vortices.
        - Node quantities are shown as colors on the nodes.
        - Arrow quantities are displayed at the junctions and the length of the arrows is
          proportional to the quantity value.
        - Face quantities are shown as colors on the faces.
        - Vortices are displayed by symbols (concentric rings, vorticity equals nr of rings,
          color shows sign).

    Base class used by :py:attr:`circuit_visualize.ConfigPlot`, where one can specify
    what physical quantities to display in the circuit.

    Parameters
    ----------
    circuit : :py:attr:`josephson_circuit.Circuit`
        Object representing circuit.
    node_data=None : (Nn,) array or None
        Data associated with nodes in circuit. Can be visualized as colors.
        If None, node data is not visualized.
    arrow_data=None : (Nj,) array or None
        Data associated with junctions in circuit. Can be visualized as arrows.
        where the length corresponds with the magnitude of the data.
        If None, arrow data is not visualized.
    face_data=None : (Nf,) array or None
        Data associated with faces in circuit. Can be visualized as colors.
        If None, face data is not visualized.
    vortex_data=None : (Nf,) int array or None
        Integer data associated with faces in circuit. Can be visualized as
        circular symbols. The value of the data equals the number of concentric
        rings of the symbol. The color shows if it is + or -.
        If None, vortex data is not visualized.
    vortex_diameter=0.25 : float
        Diameter of vortex symbols.
    vortex_color=(0, 0, 0) : color
        Color of vortex symbols.
    anti_vortex_color=(0.8, 0.1, 0.2) : color
        Color of anti-vortex symbols, whose data is negative.
    vortex_alpha=1 : float
        Transparency of vortex symbols.
    vortex_label="" : str
        Label given to vortex data in legend.
    show_grid=True : bool
        Display a grid at the edges of the graph.
    grid_width=1 : float
        Width of lines of grid.
    grid_color=(0.4, 0.5, 0.6) : color
        Color of grid.
    grid_alpha=0.5 : float
        Transparency of grid.
    show_colorbar=True : bool
        Show colorbar mapping face and/or node data to colors.
    show_legend=True : bool
        Show legend which includes colormaps, explanation of vortex sybols and
        an arrow scale.
    show_axes=True : bool
        If True, shows axes with the coordinates of the circuit.
    arrow_width=0.005 : float
        Width of arrows.
    arrow_scale=1 : float
        Scale-factor for arrows. (length of arrow = arrow_scale * arrow_data)
    arrow_headwidth=3 : float
        Width of head of arrows. (see matplotlib.quiver)
    arrow_headlength=5 : float
        Length of head of arrows. (see matplotlib.quiver)
    arrow_headaxislength=4.5 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minshaft=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minlength=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_color=(0.15, 0.3, 0.8) : color
        Color of arrows.
    arrow_alpha=1 : float
        Transparency of arrows.
    arrow_label="" : str
        Label given to arrow data in legend.
    show_nodes=True : bool
        If True, nodes are displayed as circles.
    node_diameter=0.25 : float
        Diameter of nodes.
    node_face_color=(1,1,1) : color
        Color of faces of nodes. Only used if there is no node data.
    node_edge_color=(0, 0, 0) : color
        Color of edge of nodes.  Only used if there is no node data.
    nodes_as_voronoi=False : bool
        If True, node data is visualized as colors of faces of a
        voronoi diagram based on node coordinates rather than color
        of circles at node coordinates.
    node_alpha=1 : float
        Transparency of nodes.
    node_quantity_cmap=None : colormap or None
        Colormap for node_data.
    node_quantity_clim=None : (float, float) or None
        Color limits for node_data.
    node_quantity_alpha=1 : float
        Transparency of colors used to represent node_data.
    node_quantity_logarithmic_colors=False : bool
        If True, node_data color-scale is logarithmic.
    node_label="" : str
        Label given to the node colormap.
    face_quantity_cmap=None : colormap or None
        Colormap for face_data.
    face_quantity_clim=None : (float, float) or None
        Color limits for face_data.
    face_quantity_alpha=1 : float
        Transparency of colors used to represent face_data.
    face_quantity_logarithmic_colors=False : bool
        If True, face_data color-scale is logarithmic.
    face_label="" : str
        Label given to the face colormap.
    figsize=None : (float, float) or None
        Size of figure in inches.
    title="" : str
        Title given to figure.
    """

    def __init__(self, circuit: Circuit, node_data=None,
                 arrow_data=None, face_data=None, vortex_data=None,
                 vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2),
                 vortex_alpha=1, vortex_label="",
                 show_grid=True, grid_width=1, grid_color=(0.4, 0.5, 0.6), grid_alpha=0.5,
                 show_colorbar=True, show_legend=True, show_axes=True,
                 arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1,
                 arrow_color=(0.15, 0.3, 0.8), arrow_alpha=1, arrow_label="",
                 show_nodes=True, node_diameter=0.25, node_face_color=(1,1,1),
                 node_edge_color=(0, 0, 0),
                 nodes_as_voronoi=False,
                 node_alpha=1, node_quantity_cmap=None,
                 node_quantity_clim=None, node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False, node_label="",
                 face_quantity_cmap=None, face_quantity_clim=None, face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False, face_label="",
                 figsize=None, title="", _vortex_range=None, fig=None):


        self.circuit = circuit

        self.node_data = node_data
        self.arrow_data = arrow_data
        self.face_data = face_data

        self.vortex_data = np.array(vortex_data, dtype=int) if vortex_data is not None else vortex_data

        self.show_vortices = self.vortex_data is not None
        self.vortex_diameter = vortex_diameter
        self.vortex_color = vortex_color
        self.anti_vortex_color = anti_vortex_color
        self.vortex_alpha = vortex_alpha
        self.vortex_label = vortex_label
        self._assign_vortex_range(_vortex_range)

        self.show_grid = show_grid
        self.grid_width = grid_width
        self.grid_color = grid_color
        self.grid_alpha = grid_alpha
        self.show_colorbar = show_colorbar
        self.show_legend = show_legend
        self.show_axes = show_axes

        self.show_arrows = self.arrow_data is not None
        self.arrow_width = arrow_width
        self._assign_arrow_scale(arrow_scale)
        self.arrow_headwidth = arrow_headwidth
        self.arrow_headlength = arrow_headlength
        self.arrow_headaxislength = arrow_headaxislength
        self.arrow_minshaft = arrow_minshaft
        self.arrow_minlength = arrow_minlength
        self.arrow_color = arrow_color
        self.arrow_alpha = arrow_alpha
        self.arrow_label = arrow_label

        self.show_nodes = show_nodes
        self.node_diameter = node_diameter
        self.node_face_color = node_face_color
        self.node_edge_color = node_edge_color
        self.node_alpha = node_alpha
        self.show_node_quantity = (self.node_data is not None) and show_nodes
        self.node_quantity_cmap = node_quantity_cmap
        if self.node_quantity_cmap is None:
            self.node_quantity_cmap = "twilight"
        self.node_quantity_clim = node_quantity_clim
        self.node_quantity_alpha = node_quantity_alpha
        self.node_quantity_logarithmic_colors = node_quantity_logarithmic_colors
        self.node_label = node_label
        self.nodes_as_voronoi = nodes_as_voronoi

        self.show_face_quantity = self.face_data is not None
        self.face_quantity_cmap = face_quantity_cmap
        if self.face_quantity_cmap is None:
            self.face_quantity_cmap = "pink"
        self.face_quantity_clim = face_quantity_clim
        self.face_quantity_alpha = face_quantity_alpha
        self.face_quantity_logarithmic_colors = face_quantity_logarithmic_colors
        self.face_label = face_label

        self.figsize = figsize if figsize is not None else [6.4, 5]
        self.colorbar1 = None
        self.cb1_label = None
        self.colorbar2 = None
        self.cb2_label = None
        self.title = title
        self.fig = fig
        self.ax = None
        self.ax_legend = None
        self.ax_cb1 = None
        self.ax_cb2 = None

        self.grid_handle = None
        self.node_handle = None
        self.face_handle = None
        self.arrow_handle = None
        self.vortex_handles = []

    def _prepare_legend(self, left, bottom, width, height):
        self.ax_legend = self.fig.add_axes([left, bottom, width, height])
        self.legend_stack = []
        self.legend_stack += [0.75] if self.show_arrows else [0.93]
        self.legend_stack += [0.75] if self.show_arrows else [0.93]
        s = self.show_vortices and len(self._vortex_range) > 0
        self.legend_stack += [self._vortex_legend_get_ymin(self.legend_stack[-1])[0]] if s else [self.legend_stack[-1]]
        s = max(0, self.legend_stack[-1] - 0.5)
        if self.show_node_quantity and self.show_face_quantity:
            self.ax_cb1 = self.ax_legend.inset_axes([0.01, 0.02 + s, 0.18, min(self.legend_stack[-1] - 0.1, 0.4)])
        else:   # at most one colorbar
            self.ax_cb1 = self.ax_legend.inset_axes([0.3, 0.02 + s, 0.18, min(self.legend_stack[-1] - 0.1, 0.4)])
        self.ax_cb2 = self.ax_legend.inset_axes([0.64, 0.02 + s, 0.18, min(self.legend_stack[-1] - 0.1, 0.4)])

    def make(self):
        """
        Make circuit plot.

        Returns
        -------
        fig :
            figure handle
        ax :
            axis handle
        """
        if self.fig is None:
            self.fig = plt.figure(figsize=self.figsize)
        else:
            self.fig.clear()
        left, width, l_width = 0.1, 0.68, 0.17
        bottom, height = 0.1, 0.85
        spacing = 0.005

        if self.show_legend:
            self.ax = self.fig.add_axes([left, bottom, width, height])
        else:
            self.ax = self.fig.add_axes([left, bottom, width + l_width - 0.02, height])

        if self.show_legend:
            self._prepare_legend(left + width + spacing, bottom, l_width, height)

        self.ax.set_title(self.title)
        self._set_axes()
        if not self.show_axes:
            self.ax.set_axis_off()

        # plot data
        if self.show_grid:
            self._plot_grid()
        if self.show_face_quantity:
            self._plot_faces()
        if self.show_nodes:
            self._plot_nodes()
        if self.show_arrows:
            self._plot_arrows()
        if self.show_vortices:
            self._plot_vortices()
        if self.show_legend:
            if self.colorbar1 is None:
                self.ax_cb1.set_axis_off()
            if self.colorbar2 is None:
                self.ax_cb2.set_axis_off()
            self._make_legend()

        xmin, xmax, ymin, ymax = self._get_lims()
        self.ax.set_xlim(xmin, xmax)
        self.ax.set_ylim(ymin, ymax)

        # return handles
        return self.fig, self.ax

    def get_node_data(self):
        """
        Get data visualized at nodes.
        """
        return self.node_data

    def get_arrow_data(self):
        """
        Get data visualized with arrows.
        """
        return self.arrow_data

    def get_junction_data(self):
        """
        Get data visualized with junctions.
        """
        return self.arrow_data

    def get_face_data(self):
        """
        Get data visualized at faces with colors.
        """
        return self.face_data

    def get_vortex_data(self):
        """
        Get integer data visualized at faces with symbols.
        """
        return self.vortex_data

    def get_figure_handle(self):
        """
        Returns figure handle.
        """
        return self.fig

    def get_axes_handle(self):
        """
        Returns main axes containing the visualization of the circuit.
        """
        return self.fig

    def get_legend_axis_handle(self):
        """
        Get axis-handle containing the legend (with arrow-scale, vortex symbol legend
        and colormaps).
        """
        return self.ax_legend

    def get_colorbar1_handle(self):
        """
        Get handle of first (left) colorbar. Representing node-or face data colors.
        """
        return self.colorbar1

    def get_colorbar2_handle(self):
        """
        Get handle of second (right) colorbar.
        """
        return self.colorbar2

    def get_node_data_handle(self):
        """
        Get handle containing visualization of node_data.
        """
        return self.node_handle

    def get_face_data_handle(self):
        """
        Get handle containing visualization of face_data.
        """
        return self.face_handle

    def get_vortex_data_handles(self):
        """
        Get handles containing vortex symbols.
        """
        return self.vortex_handles

    def get_arrow_data_handles(self):
        """
        Get handle containing visualization of arrow_data.
        """
        return self.arrow_handle

    def get_grid_handle(self):
        """
        Get handle containing grid.
        """
        return self.grid_handle

    # def _return_figure_handles(self):
    #     handles = [self.fig, self.ax, self.colorbar1, self.colorbar2]
    #     return [h for h in handles if h is not None]

    def _assign_arrow_scale(self, arrow_scale):
        self.arrow_scale = arrow_scale
        if self.show_arrows and self.arrow_scale is None:
            self.arrow_scale = np.max(np.abs(self.arrow_data))
        self.arrow_scale = 1 if np.isclose(self.arrow_scale, 0) else self.arrow_scale

    def _assign_vortex_range(self, vortex_range):
        self._vortex_range = vortex_range
        if vortex_range is None and self.show_vortices:
            self._vortex_range = np.unique(self.vortex_data)
        if self._vortex_range is not None:
            self._vortex_range = self._vortex_range[self._vortex_range != 0]

    def _get_lims(self):
        x, y = self.circuit.get_node_coordinates()
        xmin, xmax, ymin, ymax = np.min(x), np.max(x), np.min(y), np.max(y)
        dx, dy = xmax - xmin, ymax - ymin
        D = self.node_diameter * 0.5
        return xmin - 0.05 * dx - D, xmax + 0.05 * dx + D, ymin - 0.05 * dy - D, ymax + 0.05 * dy + D

    def _set_axes(self, ):
        xmin, xmax, ymin, ymax = self._get_lims()
        self.ax.set_xlim(xmin, xmax)
        self.ax.set_ylim(ymin, ymax)
        self.time_label = self.ax.annotate("", (xmin + 0.99 * (xmax - xmin), ymin + 0.98 * (ymax - ymin)), color=[1, 0.5, 0.2], ha='right', va='center')
        x0, y0, width, height = self.ax.get_position().bounds
        a_width = width * self.figsize[0]
        a_height = height * self.figsize[1]
        if a_width / (xmax - xmin) > a_height / (ymax - ymin):
            new_width = a_height * (xmax - xmin) / (ymax - ymin) / self.figsize[0]
            x0 = x0 + (width - new_width) / 2
            width = new_width
        else:
            new_height = a_width * (ymax - ymin) / (xmax - xmin) / self.figsize[1]
            y0 = y0 + (height - new_height) / 2
            height = new_height
        self.ax.set_position([x0, y0, width, height])

    def _marker_scale_factor(self):
        return self._scale_factor(self.ax, True) * 72

    def _scale_factor(self, ax, on_x):
        x0, y0, width, height = ax.get_position().bounds
        if on_x:
            xlim = ax.get_xlim()
            return (width * self.figsize[0]) / (xlim[1] - xlim[0])
        else:
            ylim = ax.get_ylim()
            return (height * self.figsize[1]) / (ylim[1] - ylim[0])

    def _plot_grid(self):
        x1, y1, x2, y2 = self.circuit.get_juncion_coordinates()
        lines = [((x1[i], y1[i]), (x2[i], y2[i])) for i in range(len(x1))]
        lc = LineCollection(lines, colors=self.grid_color,  alpha=self.grid_alpha,linewidths=self.grid_width, zorder=0)
        self.grid_handle = self.ax.add_collection(lc)

    @staticmethod
    def _voronoi(ax, x, y, data, cnorm, cmap, alpha):
        points = np.stack((x.flatten(), y.flatten()), axis=0).T
        xm, xp = ax.get_xlim()
        ym, yp = ax.get_ylim()
        points = np.append(points, [[xm - 1, ym - 1], [xm - 1, yp + 1], [xp + 1, yp + 1], [xp + 1, ym - 1]], axis=0)
        vor = Voronoi(points)
        verts = [vor.vertices[region, :] for region in vor.regions]
        verts = [verts[i] for i in vor.point_region[:len(x)]]
        coll = PolyCollection(verts, array=data, edgecolors='none', cmap=cmap,
                              norm=cnorm, alpha=alpha, zorder=-1)
        return ax.add_collection(coll)

    def _cnorm(self, clim, data, logarithmic):
        if clim is None:
            clim = (np.min(data), np.max(data))
        return clim, Normalize(*clim) if not logarithmic else LogNorm(*clim)

    def _assign_colorbar(self, cmap, clim, label, cb_nr):
        if cb_nr == 0:
            self.colorbar1 = matplotlib.colorbar.Colorbar(ax=self.ax_cb1, mappable=cmap)
            self.cb1_label = label
            if np.allclose(clim, [-np.pi, np.pi]):
                self._set_ax_pi_ticks(self.colorbar1)
            else:
                self.ax_cb1.set_ylim(clim)
        else:
            self.colorbar2 = matplotlib.colorbar.Colorbar(ax=self.ax_cb2, mappable=cmap)
            self.cb2_label = label
            if np.allclose(clim, [-np.pi, np.pi]):
                self._set_ax_pi_ticks(self.colorbar2)
            else:
                self.ax_cb2.set_ylim(clim)

    def _plot_nodes(self):
        x, y = self.circuit.get_node_coordinates()

        marker_size = self.node_diameter * self._marker_scale_factor()
        if not self.show_node_quantity:
            node_handle = self.ax.plot(x, y, markeredgecolor=self.node_edge_color,
                                       markerfacecolor=self.node_face_color,
                                       linestyle="None",
                                       markersize=marker_size, marker="o",
                                       alpha=self.node_alpha, zorder=2)
            self.node_handle = node_handle[0]
        else:
            clim, cnorm = self._cnorm(self.node_quantity_clim, self.node_data, self.node_quantity_logarithmic_colors)
            if self.nodes_as_voronoi:
                self.node_handle = self._voronoi(self.ax, x, y, self.node_data, cnorm,
                                                 self.node_quantity_cmap, self.node_quantity_alpha)
            else:
                self.node_handle = self.ax.scatter(x.flatten(), y.flatten(), s=marker_size**2,
                                                   c=self.node_data, cmap=self.node_quantity_cmap,
                                                   edgecolors=None,
                                                   alpha=self.node_quantity_alpha, norm=cnorm)
            if self.show_colorbar and self.show_legend:
                cmap = cm.ScalarMappable(norm=cnorm, cmap=self.node_quantity_cmap)
                if self.colorbar1 is None:
                    self._assign_colorbar(cmap, clim, self.node_label, cb_nr=0)
                else:
                    self._assign_colorbar(cmap, clim, self.node_label, cb_nr=1)

    def _set_ax_pi_ticks(self, cb):
        cb.set_ticks([-np.pi, -0.5 * np.pi, 0, 0.5*np.pi, np.pi])
        cb.set_ticklabels(["$-\pi$", "$-\pi/2$", "$0$", "$\pi/2$", "$\pi$"])

    def _plot_arrows(self):
        I = self.arrow_data * self.arrow_scale
        x1, y1, x2, y2 = self.circuit.get_juncion_coordinates()
        xq = x1 + 0.5 * (1 - I) * (x2 - x1)
        yq = y1 + 0.5 * (1 - I) * (y2 - y1)
        dxq, dyq = I * (x2 - x1), I * (y2 - y1)
        self.arrow_handle = self.ax.quiver(xq, yq, dxq, dyq, edgecolor=self.arrow_color,
                                           facecolor=self.arrow_color, angles='xy', scale=1,
                                           scale_units='xy', width=self.arrow_width,
                                           headwidth=self.arrow_headwidth, headlength=self.arrow_headlength,
                                           headaxislength=self.arrow_headaxislength, minshaft=self.arrow_minshaft,
                                           minlength=self.arrow_minlength, alpha=self.arrow_alpha, zorder=3)

    def _plot_faces(self):
        face_nodes = self.circuit.get_faces()
        x, y = self.circuit.get_node_coordinates()
        verts = [np.stack((x[n], y[n]), axis=-1) for n in face_nodes]
        clim = self.face_quantity_clim
        if clim is None:
            clim = (np.min(self.face_data), np.max(self.face_data))
        cnorm = Normalize(*clim) if not self.face_quantity_logarithmic_colors else LogNorm(*clim)
        coll = PolyCollection(verts, array=self.face_data, edgecolors='none',
                              cmap=self.face_quantity_cmap,
                              norm=cnorm, alpha=self.face_quantity_alpha, zorder=-1)
        self.face_handle = self.ax.add_collection(coll)
        if self.show_colorbar and self.show_legend:
            cmap = cm.ScalarMappable(norm=cnorm, cmap=self.face_quantity_cmap)
            self._assign_colorbar(cmap, clim, self.face_label, cb_nr=0)

    def _plot_vortices(self):
        n = self.vortex_data
        marker_size = self.vortex_diameter * self._marker_scale_factor()
        xc, yc = self.circuit.get_face_centroids()
        ns = self._vortex_range
        vort_handles = []
        for ni in ns:
            n_mask = n == ni
            xni, yni = xc[n_mask], yc[n_mask]
            p = self._draw_vortex(self.ax, xni, yni, ni, self.vortex_color,
                                  self.anti_vortex_color, marker_size, self.vortex_alpha)
            vort_handles += p
        self.vortex_handles = vort_handles

    def _update_nodes(self, new_node_data):
        if new_node_data is not None and self.show_node_quantity:
            self.node_handle.set_array(new_node_data)

    def _update_arrows(self, new_arrow_data):
        if new_arrow_data is not None and self.show_arrows:
            I = new_arrow_data * self.arrow_scale
            x1, y1, x2, y2 = self.circuit.get_juncion_coordinates()
            xq, yq = x1 + 0.5 * (1 - I) * (x2 - x1), y1 + 0.5 * (1 - I) * (y2 - y1)
            U, V = I * (x2 - x1), I * (y2 - y1)
            self.arrow_handle.set_UVC(U, V)
            self.arrow_handle.X = xq
            self.arrow_handle.Y = yq
            self.arrow_handle.XY = np.concatenate((xq[:, None], yq[:, None]), axis=1)
            self.arrow_handle._offsets = self.arrow_handle.XY

    def _update_faces(self, new_face_data):
        if new_face_data is not None and self.show_face_quantity:
            self.face_handle.set_array(new_face_data)

    def _update_vortices(self, new_vortex_data):
        if new_vortex_data is not None and self.show_vortices:
            n = new_vortex_data
            xc, yc = self.circuit.get_face_centroids()
            ns = self._vortex_range
            index = 0
            for i, ni in enumerate(ns):
                na = np.abs(ni)
                for handle in self.vortex_handles[index:(index + 2 * na)]:
                    handle.set_xdata(xc[n == ni])
                    handle.set_ydata(yc[n == ni])
                index += 2 * na - 1

    def _make_legend(self):
        self.ax_legend.set_xlim([0, 1])
        self.ax_legend.set_ylim([0, 1])
        self.ax_legend.set_axis_off()

        # arrow legend
        self.ax_legend.text(0.5, 0.975, "legend", ha="center", va="center", fontsize=13)
        if self.show_arrows:
            self._arrow_legend(0.75, 0.93)

        # vortex legend
        if self.show_vortices and len(self._vortex_range) > 0:
            self._vortex_legend(self.legend_stack[0])

        # node/face legend
        if self.cb1_label is not None:
            if self.cb2_label is not None:
                self.ax_legend.text(0.11, self.legend_stack[-1] - 0.05, self.cb1_label, ha="center", va="center")
            else:
                self.ax_legend.text(0.4, self.legend_stack[-1] - 0.05, self.cb1_label, ha="center", va="center")
        if self.cb2_label is not None:
            self.ax_legend.text(0.75, self.legend_stack[-1] - 0.05, self.cb2_label, ha="center", va="center")

    @staticmethod
    def _nearest_anchor(x):
        b = np.log(x)/np.log(10)
        decade = np.floor(b)
        B = b - decade
        s = np.array([0, np.log(2)/np.log(10), np.log(5)/np.log(10), 1])
        sub = np.argmin(np.abs(B - s))
        if sub == 0:
            return 10 ** decade
        if sub == 1:
            return 2 * 10 ** decade
        if sub == 2:
            return 5 * 10 ** decade
        if sub == 3:
            return 10 ** (decade + 1)

    def _legend_arrow_scale(self):
        _, _, width, _ = self.ax.get_position().bounds
        _, _, width_L, _ = self.ax_legend.get_position().bounds
        xlim = self.ax.get_xlim()
        xlim_L = self.ax_legend.get_xlim()
        v = np.median(self.circuit._junction_lengths())
        dx = xlim[1] - xlim[0]
        dx_L = xlim_L[1] - xlim_L[0]
        return v * (width / dx) / (width_L / dx_L), v / self.arrow_scale

    def _arrow_legend(self, y_min, y_max):
        L, label = self._legend_arrow_scale()
        label *= (0.51 / L)
        L = 0.51
        new_label = self._nearest_anchor(label)
        L *= new_label / label
        ax =self.ax_legend
        y = 0.6 * y_min + 0.4 * y_max
        xc = 0.58
        hL = L/2
        ax.text(0.12, y, self.arrow_label, ha="center", va="center")
        ax.quiver([xc-hL], [y], [L], [0], edgecolor=self.arrow_color, facecolor=self.arrow_color,
                  scale_units="xy", angles="xy", scale=1, width=0.02)
        dy = 0.01
        Y = 0.4 * y_min + 0.6 * y_max
        ax.plot([xc-hL, xc-hL, xc-hL, xc+hL, xc+hL, xc+hL], [Y-dy, Y+dy, Y, Y, Y+dy, Y-dy], color=[0, 0, 0])
        yy = 0.2 * y_min + 0.8 * y_max
        ax.text(xc, yy, f"{new_label}", ha="center", va="center")

    def _vortex_legend_get_ymin(self, y_max):
        n_range= self._vortex_range
        y_min = -np.inf
        scale_factor_x = self._scale_factor(self.ax_legend, True)
        scale_factor_y = self._scale_factor(self.ax_legend, False)
        d = 0.3
        while y_min < 0.25:
            d *= 0.8
            Dy = 1.1 * d * (len(n_range) + 1) / (scale_factor_y / scale_factor_x)
            y_min = y_max - Dy
        return y_min, d

    def _vortex_legend(self, y_max):
        ax = self.ax_legend
        n_range= self._vortex_range
        v_color, av_color = self.vortex_color, self.anti_vortex_color
        label = self.vortex_label
        scale_factor_x = self._scale_factor(ax, True)
        y_min, d = self._vortex_legend_get_ymin(y_max)
        Y = np.linspace(y_max, y_min, 2 * len(n_range) + 2)
        y = Y[2:-1:2]
        ax.text(0.12, Y[1], label, ha="center", va="bottom")
        ax.text(0.6, Y[1], "symbol", ha="center", va="bottom")
        for i, n in enumerate(n_range):
            CircuitPlot._draw_vortex(ax, 0.6, y[i], n, v_color, av_color, scale_factor_x * d * 72, 1)
            ax.text(0.12, y[i], f"{n}", va="center", ha="center")
        plt.plot([0.02, 0.99, 0.99, 0.02, 0.02], [y_min, y_min, y_max+0.016, y_max+0.016, y_min], color=[0, 0, 0])
        return y_min

    @staticmethod
    def _draw_vortex(ax, x, y, n, v_color, av_color, diameter, alpha):
        color = v_color if n > 0 else av_color
        na = np.abs(n)
        handles = []
        for k in reversed(range(na)):
            frac = (2 * k + 1) / (2 * na - 1)
            p = ax.plot(x, y, markerfacecolor=color, markeredgecolor=color, marker="o",
                        linestyle="", markersize=frac * diameter, alpha=alpha, zorder=4)
            handles += p
            if k > 0:
                frac = (2 * k) / (2 * na - 1)
                p = ax.plot(x, y, markerfacecolor=[1, 1, 1], markeredgecolor=[1, 1, 1], marker="o",
                            linestyle="", markersize=frac * diameter, alpha=alpha, zorder=4)
                handles += p
        return handles


class CircuitMovie(CircuitPlot):
    """Animation of circuit where the quantities evolve over time.

    Allows one to show node quantities, arrow quantities, face quantities and vortices.
        - Node quantities are shown as colors on the nodes.
        - Arrow quantities are displayed at the junctions and the length of the arrows is
          proportional to the quantity value.
        - Face quantities are shown as colors on the faces.
        - Vortices are displayed by symbols (concentric rings, vorticity equals nr of rings,
          color shows sign).

    Base class used by :py:attr:`circuit_visualize.TimeEvolutionMovie`, where one can
    specify what physical quantities to display in the circuit.

    Parameters
    ----------
    circuit : :py:attr:`josephson_circuit.Circuit`
        Object representing circuit.
    animate_interval : scalar
        Frame delay in ms.
    node_data=None : (Nn, time_points) array or (Nn,) array or None
        Data associated with nodes in circuit. Can be visualized as colors.
        If None, node data is not visualized. Time independent if shape (Nn,).
    arrow_data=None : (Nj, time_points) array or (Nj,) array or None
        Data associated with junctions in circuit. Can be visualized as arrows.
        where the length corresponds with the magnitude of the data.
        If None, arrow data is not visualized. Time independent if shape (Nj,).
    face_data=None : (Nf, time_points) array or (Nf,) array or None
        Data associated with faces in circuit. Can be visualized as colors.
        If None, face data is not visualized. Time independent if shape (Nf,).
    vortex_data=None : (Nf, time_points) int array or (Nf,) int array or None
        Integer data associated with faces in circuit. Can be visualized as
        circular symbols. The value of the data equals the number of concentric
        rings of the symbol. The color shows if it is + or -.
        If None, vortex data is not visualized. Time independent if shape (Nf,).
    vortex_diameter=0.25 : float
        Diameter of vortex symbols.
    vortex_color=(0, 0, 0) : color
        Color of vortex symbols.
    anti_vortex_color=(0.8, 0.1, 0.2) : color
        Color of anti-vortex symbols, whose data is negative.
    vortex_alpha=1 : float
        Transparency of vortex symbols.
    vortex_label="" : str
        Label given to vortex data in legend.
    show_grid=True : bool
        Display a grid at the edges of the graph.
    grid_width=1 : float
        Width of lines of grid.
    grid_color=(0.4, 0.5, 0.6) : color
        Color of grid.
    grid_alpha=0.5 : float
        Transparency of grid.
    show_colorbar=True : bool
        Show colorbar mapping face and/or node data to colors.
    show_legend=True : bool
        Show legend which includes colormaps, explanation of vortex sybols and
        an arrow scale.
    show_axes=True : bool
        If True, shows axes with the coordinates of the circuit.
    arrow_width=0.005 : float
        Width of arrows.
    arrow_scale=1 : float
        Scale-factor for arrows. (length of arrow = arrow_scale * arrow_data)
    arrow_headwidth=3 : float
        Width of head of arrows. (see matplotlib.quiver)
    arrow_headlength=5 : float
        Length of head of arrows. (see matplotlib.quiver)
    arrow_headaxislength=4.5 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minshaft=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minlength=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_color=(0.15, 0.3, 0.8) : color
        Color of arrows.
    arrow_alpha=1 : float
        Transparency of arrows.
    arrow_label="" : str
        Label given to arrow data in legend.
    show_nodes=True : bool
        If True, nodes are displayed as circles.
    node_diameter=0.25 : float
        Diameter of nodes.
    node_face_color=(1,1,1) : color
        Color of faces of nodes. Only used if there is no node data.
    node_edge_color=(0, 0, 0) : color
        Color of edge of nodes.  Only used if there is no node data.
    nodes_as_voronoi=False : bool
        If True, node data is visualized as colors of faces of a
        voronoi diagram based on node coordinates rather than color
        of circles at node coordinates.
    node_alpha=1 : float
        Transparency of nodes.
    node_quantity_cmap=None : colormap or None
        Colormap for node_data.
    node_quantity_clim=None : (float, float) or None
        Color limits for node_data.
    node_quantity_alpha=1 : float
        Transparency of colors used to represent node_data.
    node_quantity_logarithmic_colors=False : bool
        If True, node_data color-scale is logarithmic.
    node_label="" : str
        Label given to the node colormap.
    face_quantity_cmap=None : colormap or None
        Volormap for face_data
    face_quantity_clim=None : (float, float) or None
        Color limits for face_data.
    face_quantity_alpha=1 : float
        Transparency of colors used to represent face_data.
    face_quantity_logarithmic_colors=False : bool
        If True, face_data color-scale is logarithmic.
    face_label="" : str
        Label given to the face colormap.
    figsize=None : (float, float) or None
        Size of figure in inches.
    title="" : str
        Title given to figure.
    """

    def __init__(self, circuit: Circuit, animate_interval=10,
                 node_data=None, arrow_data=None, face_data=None, vortex_data=None,
                 vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2),
                 vortex_alpha=1, vortex_label="", _vortex_range=None,
                 show_grid=True, grid_width=1, grid_color=(0.4, 0.5, 0.6), grid_alpha=0.5,
                 show_colorbar=True, show_legend=True, show_axes=True,
                 arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1,
                 arrow_color=(0.15, 0.3, 0.8), arrow_alpha=1, arrow_label="",
                 show_nodes=True, node_diameter=0.25, node_face_color=(1,1,1),
                 node_edge_color=(0, 0, 0),
                 nodes_as_voronoi=False,
                 node_alpha=1, node_quantity_cmap=None,
                 node_quantity_clim=None, node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False, node_label="",
                 face_quantity_cmap=None, face_quantity_clim=None, face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False, face_label="",
                 figsize=None, title="", fig=None):

        def handle_data(data, N):
            if data is None:
                return None, 1
            data = np.array(data)
            if data.ndim == 1:
                data = data[:, None]
            if data.ndim != 2 or data.shape[0] != N:
                raise ValueError(f"data must None, of shape ({N},), ({N}, 1) or ({N}, Nt)")
            Nt = data.shape[1]
            return data, Nt

        self.all_node_data, Ntn = handle_data(node_data, circuit.node_count())
        self.all_arrow_data, Nta = handle_data(arrow_data, circuit.junction_count())
        self.all_face_data, Ntf = handle_data(face_data, circuit.face_count())
        self.all_vortex_data, Ntv = handle_data(vortex_data, circuit.face_count())
        Nts = np.array([Ntn, Nta, Ntf, Ntv], dtype=int)
        self.Nt = np.max(Nts)
        if not np.all((Nts == 1) | (Nts == self.Nt)):
            raise ValueError("Not all data has same number of timesteps")

        if self.all_node_data is not None:
            if node_quantity_clim is None:
                node_quantity_clim = (np.min(self.all_node_data), np.max(self.all_node_data))
        if self.all_face_data is not None:
            if face_quantity_clim is None:
                face_quantity_clim = (np.min(self.all_face_data), np.max(self.all_face_data))

        super().__init__(circuit,  node_data=self.all_node_data[:, 0] if self.all_node_data is not None else None,
                         arrow_data=self.all_arrow_data[:, 0] if self.all_arrow_data is not None else None,
                         face_data=self.all_face_data[:, 0] if self.all_face_data is not None else None,
                         vortex_data=self.all_vortex_data[:, 0] if self.all_vortex_data is not None else None,
                         vortex_label=vortex_label, node_label=node_label, arrow_label=arrow_label, face_label=face_label,
                         vortex_diameter=vortex_diameter, vortex_color=vortex_color,
                         anti_vortex_color=anti_vortex_color,
                         vortex_alpha=vortex_alpha,
                         show_grid=show_grid, grid_width=grid_width, grid_color=grid_color,
                         grid_alpha=grid_alpha, show_colorbar=show_colorbar, show_legend=show_legend,
                         show_axes=show_axes, arrow_width=arrow_width, arrow_scale=arrow_scale,
                         arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                         arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft, arrow_minlength=arrow_minlength,
                         arrow_color=arrow_color, arrow_alpha=arrow_alpha, show_nodes=show_nodes,
                         node_diameter=node_diameter, node_face_color=node_face_color, node_edge_color=node_edge_color,
                         node_alpha=node_alpha, nodes_as_voronoi=nodes_as_voronoi,
                         node_quantity_cmap=node_quantity_cmap, node_quantity_clim=node_quantity_clim,
                         node_quantity_alpha=node_quantity_alpha, node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                         face_quantity_cmap=face_quantity_cmap,
                         face_quantity_clim=face_quantity_clim, face_quantity_alpha=face_quantity_alpha,
                         face_quantity_logarithmic_colors=face_quantity_logarithmic_colors, figsize=figsize,
                         title=title, fig=fig)

        self.animate_interval = animate_interval

    def _init(self):
        handles = [self.face_handle, self.node_handle, self.arrow_handle] + self.vortex_handles
        return [h for h in handles if h is not None]

    def _animate(self, i):
        if self.show_face_quantity:
            if self.all_face_data.shape[1] > 1:
                self._update_faces(self.all_face_data[:, i])
        if self.show_node_quantity:
            if self.all_node_data.shape[1] > 1:
                self._update_nodes(self.all_node_data[:, i])
        if self.show_arrows:
            if self.all_arrow_data.shape[1] > 1:
                self._update_arrows(self.all_arrow_data[:, i])
        if self.show_vortices:
            if self.all_vortex_data.shape[1] > 1:
                self._update_vortices(self.all_vortex_data[:, i])
        handles = [self.node_handle, self.arrow_handle, self.face_handle] + self.vortex_handles
        return [h for h in handles if h is not None]

    def make(self):
        """
        Make circuit plot.

        Returns
        -------
        fig :
            Figure handle.
        ax :
            Axis handle.
        """
        super().make()
        self.ani = animation.FuncAnimation(self.fig, self._animate, np.arange(self.Nt, dtype=int),
                                           init_func=self._init, interval=self.animate_interval,
                                           blit=True)
        return self.fig, self.ax

    def get_animation_handle(self):
        """
        Return FuncAnimation handle.
        """
        return self.ani

    def _assign_arrow_scale(self, arrow_scale):
        self.arrow_scale = arrow_scale
        if self.show_arrows and self.arrow_scale is None:
            self.arrow_scale = np.max(np.abs(self.all_arrow_data))
        self.arrow_scale = 1 if np.isclose(self.arrow_scale, 0) else self.arrow_scale

    def _assign_vortex_range(self, vortex_range):
        self._vortex_range = vortex_range
        if vortex_range is None and self.show_vortices:
            self._vortex_range = np.unique(self.all_vortex_data)
        if self._vortex_range is not None:
            self._vortex_range = self._vortex_range[self._vortex_range != 0]


class ConfigPlot(CircuitPlot):

    """
    Visualize static configuration on a circuit.

    Allows one to show node quantities, junction quantities, face quantities and vortices.
        - Node quantities are shown as colors on the nodes.
        - Junction quantities are displayed at the junctions as arrows, where the length of
          the arrows is proportional to the quantity value.
        - Face quantities are shown as colors on the faces.
        - Vortices are displayed by symbols (concentric rings, vorticity equals nr of rings,
          color shows sign).

    Parameters
    ----------
    config : :py:attr:`static_problem.StaticConfiguration`
        Static configuration to plot.
    node_quantity=None : str or None
        What physical quantity of config to visualize at nodes. Options:
            * "" or None :   no quantity displayed
            * "phi" :        gauge dependent phases
            * "Is_node" :    current sourced at nodes
    junction_quantity="I" : str or None
        What physical quantity of config to visualize on junctions. Options:
            * "" or None :   no quantity displayed
            * "theta" :  gauge invariant phase difference
            * "I" :      current
            * "Is" :     junction current sources
            * "EJ" :     josephson energy
            * "EM" :     magnetic energy
            * "Etot" :   total energy
    face_quantity=None : str or None
        What physical quantity of config to visualize at faces. Options:
            * "" or None :   no quantity displayed
            * "flux" :   magnetic flux through face
            * "J" :      cycle-current
            * "n" :      vorticity
    vortex_quantity="n" : str or None
        What face-integer physical quantity of config to visualize with vortex symbols. Options:
            * "" or None : no quantity displayed
            * "n" :   vortices
    vortex_diameter=0.25 : float
        Diameter of vortex symbols.
    vortex_color=(0, 0, 0) : color
        Color of vortex symbols.
    anti_vortex_color=(0.8, 0.1, 0.2) : color
        Color of anti-vortex symbols, whose data is negative.
    vortex_alpha=1 : float
        Transparency of vortex symbols.
    show_grid=True : bool
        Display a grid at the edges of the graph.
    grid_width=1 : float
        Width of lines of grid.
    grid_color=(0.4, 0.5, 0.6) : color
        Color of grid.
    grid_alpha=0.5 : float
        Transparency of grid.
    show_colorbar=True : bool
        Show colorbar mapping face and/or node data to colors.
    show_legend=True : bool
        Show legend which includes colormaps, explanation of vortex sybols and
        an arrow scale.
    show_axes=True : bool
        If True, shows axes with the coordinates of the circuit.
    arrow_width=0.005 : float
        Width of arrows.
    arrow_scale=1 : float
        Scale-factor for arrows. (length of arrow = arrow_scale * arrow_data)
    arrow_headwidth=3 : float
        Width of head of arrows. (see matplotlib.quiver)
    arrow_headlength=5 : float
        Length of head of arrows. (see matplotlib.quiver)
    arrow_headaxislength=4.5 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minshaft=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minlength=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_color=(0.15, 0.3, 0.8) : color
        Color of arrows.
    arrow_alpha=1 : float
        Transparency of arrows.
    show_nodes=True : bool
        If True, nodes are displayed as circles.
    node_diameter=0.25 : float
        Diameter of nodes.
    node_face_color=(1,1,1) : color
        Color of faces of nodes. Only used if there is no node data.
    node_edge_color=(0, 0, 0) : color
        Color of edge of nodes.  Only used if there is no node data.
    nodes_as_voronoi=False : bool
        If True, node data is visualized as colors of faces of a
        voronoi diagram based on node coordinates rather than color
        of circles at node coordinates.
    node_alpha=1 : float
        Transparency of nodes.
    node_quantity_cmap=None : colormap or None
        Colormap for node_data
    node_quantity_clim=None : (float, float) or None
        Color limits for node_data.
    node_quantity_alpha=1 : float
        Transparency of colors used to represent node_data.
    node_quantity_logarithmic_colors=False : bool
        If True, node_data color-scale is logarithmic.
    face_quantity_cmap=None : colormap or None
        Colormap for face_data.
    face_quantity_clim=None : (float, float) or None
        Color limits for face_data.
    face_quantity_alpha=1 : float
        Transparency of colors used to represent face_data.
    face_quantity_logarithmic_colors=False : bool
        If True, face_data color-scale is logarithmic.
    figsize=None : (float, float) or None
        Size of figure in inches.
    title="" : str
        Title given to figure.

    """

    def __init__(self, config: StaticConfiguration, node_quantity=None,
                 junction_quantity="I", face_quantity=None, vortex_quantity="n",
                 vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2),
                 vortex_alpha=1, _vortex_range=None,
                 show_grid=True, grid_width=1, grid_color=(0.4, 0.5, 0.6), grid_alpha=0.5,
                 show_colorbar=True, show_legend=True, show_axes=True,
                 arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1,
                 arrow_color=(0.15, 0.3, 0.8), arrow_alpha=1,
                 show_nodes=True, node_diameter=0.25, node_face_color=(1,1,1),
                 node_edge_color=(0, 0, 0), nodes_as_voronoi=False, node_alpha=1,
                 node_quantity_cmap=None,
                 node_quantity_clim=None, node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False,
                 face_quantity_cmap=None, face_quantity_clim=None, face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False,
                 figsize=None, title="", fig=None):

        self.config = config
        self.node_quantity = node_quantity if node_quantity is not None else ""
        self.arrow_quantity = junction_quantity if junction_quantity is not None else ""
        self.face_quantity = face_quantity if face_quantity is not None else ""
        self.vortex_quantity = vortex_quantity if vortex_quantity is not None else ""

        node_data, node_label = self._get_node_quantity()
        arrow_data, arrow_label = self._get_junction_quantity()
        face_data, face_label = self._get_face_quantity()
        vortex_data, vortex_label = self._get_vortex_quantity()

        if node_label == "phi":
            if node_quantity_clim is None:
                node_quantity_clim = (-np.pi, np.pi)

        if arrow_label == "th":
            if arrow_scale is None:
                arrow_scale = 1/np.pi

        super().__init__(config.get_circuit(),  node_data = node_data, arrow_data=arrow_data, face_data=face_data, vortex_data=vortex_data,
                         vortex_label=vortex_label, node_label=node_label, arrow_label=arrow_label, face_label=face_label,
                         vortex_diameter=vortex_diameter, vortex_color=vortex_color,
                         anti_vortex_color=anti_vortex_color,
                         vortex_alpha=vortex_alpha, show_grid=show_grid, grid_width=grid_width, grid_color=grid_color,
                         grid_alpha=grid_alpha, show_colorbar=show_colorbar, show_legend=show_legend,
                         show_axes=show_axes, arrow_width=arrow_width, arrow_scale=arrow_scale,
                         arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                         arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft, arrow_minlength=arrow_minlength,
                         arrow_color=arrow_color, arrow_alpha=arrow_alpha, show_nodes=show_nodes,
                         node_diameter=node_diameter, node_face_color=node_face_color, node_edge_color=node_edge_color,
                         node_alpha=node_alpha, nodes_as_voronoi=nodes_as_voronoi,
                         node_quantity_cmap=node_quantity_cmap, node_quantity_clim=node_quantity_clim,
                         node_quantity_alpha=node_quantity_alpha, node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                         face_quantity_cmap=face_quantity_cmap,
                         face_quantity_clim=face_quantity_clim, face_quantity_alpha=face_quantity_alpha,
                         face_quantity_logarithmic_colors=face_quantity_logarithmic_colors,
                         figsize=figsize, title=title, fig=fig)


    _node_quantities = {
        "": -1,
        "phi": 0, "phase": 0, "phases": 0,
        "I_s": 1, "Is": 1, "current_sources": 1,
    }

    _junction_quantities = {
        "": -1,
        "th": 0, "theta": 0, "phase_difference": 0, "gauge_invariant_phase_difference": 0,
        "I": 1, "current": 1, "Isup": 1, "I_sup": 1, "Isuper": 1, "I_super": 1,
        "supercurrent": 1, "super_current": 1,
        "I_s": 2, "Is": 2, "current_sources": 2,
        "EJ": 3, "josephson_energy": 3, "Ej": 3,
        "EM": 4, "magnetic_energy": 4, "Em": 4,
        "Etot": 5, "E_tot": 5, "ETot": 5, "total_energy": 5, "energy": 5,
    }

    _face_quantities = {
        "": -1,
        "Phi": 0, "flux": 0, "magnetic_flux": 0,
        "n": 1, "vortices": 1, "vortex_configuration": 1,
        "face_current": 2, "J": 2, "cycle_current": 2,
    }

    _vortex_quantities = {
        "": -1,
        "n": 0, "vortices": 0, "vortex_configuration": 0,
    }

    def get_node_quantity(self):
        """
        Get physical quantity displayed at nodes. (None, "phi" or "Is_node")
        """
        return self.node_label

    def get_junction_quantity(self):
        """
        Get physical quantity displayed at junctions. (None, "th", "I",
        "Is", "EJ", "EM" or "Etot")
        """
        return self.arrow_label

    def get_face(self):
        """
        Get physical quantity displayed at faces with colors. (None, "flux", " n" or "J")
        """
        return self.face_label

    def get_vortex_quantity(self):
        """
        Get integer physical quantity displayed at faces with symbols. (None or "n")
        """
        return self.vortex_label

    def _get_node_quantity(self):
        if isinstance(self.node_quantity, np.ndarray):
            return self.node_quantity.flatten(), "custom"
        quantity = self._node_quantities[self.node_quantity]
        if quantity == -1:   # none
            return None, None
        if quantity == 0:  # phi
            out = self.config.get_phi()
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out, "phi"
        if quantity == 1:   # Is
            return self.config.problem.get_node_current_sources(), "Is_node"

    def _get_junction_quantity(self):
        if isinstance(self.arrow_quantity, np.ndarray):
            return self.arrow_quantity.flatten(), "custom"
        quantity = self._junction_quantities[self.arrow_quantity]
        if quantity == -1:   # none
            return None, None
        if quantity == 0:  # theta
            out = self.config.get_theta()
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out, "th"
        if quantity == 1:  # I
            return self.config.get_I(), "I"
        if quantity == 2:  # Is
            return self.config.problem._Is(), "Is"
        if quantity == 3:  # EJ
            return self.config.get_EJ(), "EJ"
        if quantity == 4:  # EM
            return self.config.get_EM(), "EM"
        if quantity == 5:  # Etot
            return self.config.get_Etot(), "Etot"

    def _get_face_quantity(self):
        if isinstance(self.face_quantity, np.ndarray):
            return self.face_quantity.flatten(), "custom"
        quantity = self._face_quantities[self.face_quantity]
        if quantity == -1:   # none
            return None, None
        if quantity == 0:  # Phi
            return self.config.get_flux(), "flux"
        if quantity == 1:  # n
            return self.config.get_n(), "n"
        if quantity == 2:  # J
            return self.config.get_J(), "J"

    def _get_vortex_quantity(self):
        if isinstance(self.vortex_quantity, np.ndarray):
            return self.vortex_quantity.flatten().astype(int), "custom"
        quantity = self._vortex_quantities[self.vortex_quantity]
        if quantity == -1:   # none
            return None, None
        if quantity == 0:  # n
            return self.config.get_n(), "n"

    @staticmethod
    def _assert_single_configuration(data):
        if data.ndim >= 2:
            raise ValueError("must select single configuration")


class TimeEvolutionMovie(CircuitMovie):


    """
    Visualize time evolution on a circuit as a movie.

    Allows one to show node quantities, junction quantities, face quantities and vortices.
        - Node quantities are shown as colors on the nodes.
        - Junction quantities are displayed at the junctions as arrows, where the length of
          the arrows is proportional to the quantity value.
        - Face quantities are shown as colors on the faces.
        - Vortices are displayed by symbols (concentric rings, vorticity equals nr of rings,
          color shows sign).

    Parameters
    ----------
    config : :py:attr:`time_evolution.TimeEvolutionResult`
        Time evolution to visualize.
    node_quantity=None : str or None
        What physical quantity of config to visualize at nodes. Options:
            * "" or None :   no quantity displayed
            * "phi" :        gauge dependent phases
            * "Is_node" :    current sourced at nodes
            * "U" :          Voltage potential
    junction_quantity="I" : str or None
        What physical quantity of config to visualize on junctions. Options:
            * "" or None :   no quantity displayed
            * "theta" :  gauge invariant phase difference
            * "I" :      current
            * "V" :      Voltage
            * "Isuper" : supercurrent
            * "Is" :     junction current sources
            * "EJ" :     josephson energy
            * "EM" :     magnetic energy
            * "EC" :     capacative energy
            * "Etot" :   total energy
    face_quantity=None : str or None
        What physical quantity of config to visualize at faces. Options:
            * "" or None :   no quantity displayed
            * "flux" :   magnetic flux through face
            * "J" :      cycle-current
            * "n" :      vorticity
    vortex_quantity="n" : str or None
        What face-integer physical quantity of config to visualize with vortex symbols. Options:
            * "" or None  : no quantity displayed
            * "n" :   vortices
    vortex_diameter=0.25 : float
        Diameter of vortex symbols.
    vortex_color=(0, 0, 0) : color
        Color of vortex symbols.
    anti_vortex_color=(0.8, 0.1, 0.2) : color
        Color of anti-vortex symbols, whose data is negative.
    vortex_alpha=1 : float
        Transparancy of vortex symbols.
    show_grid=True : bool
        Display a grid at the edges of the graph.
    grid_width=1 : float
        Width of lines of grid.
    grid_color=(0.4, 0.5, 0.6) : color
        Color of grid.
    grid_alpha=0.5 : float
        Transparency of grid.
    show_colorbar=True : bool
        Show colorbar mapping face and/or node data to colors.
    show_legend=True : bool
        Show legend which includes colormaps, explanation of vortex sybols and
        an arrow scale.
    show_axes=True : bool
        If True, shows axes with the coordinates of the circuit.
    arrow_width=0.005 : float
        Width of arrows.
    arrow_scale=1 : float
        Scale-factor for arrows. (length of arrow = arrow_scale * arrow_data)
    arrow_headwidth=3 : float
        Width of head of arrows. (see matplotlib.quiver)
    arrow_headlength=5 : float
        Length of head of arrows. (see matplotlib.quiver)
    arrow_headaxislength=4.5 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minshaft=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_minlength=1 : float
        Arrow property. (see matplotlib.quiver)
    arrow_color=(0.15, 0.3, 0.8) : color
        Color of arrows.
    arrow_alpha=1 : float
        Transparency of arrows.
    show_nodes=True : bool
        If True, nodes are displayed as circles
    node_diameter=0.25 : float
        Diameter of nodes.
    node_face_color=(1,1,1) : color
        Color of faces of nodes. Only used if there is no node data.
    node_edge_color=(0, 0, 0) : color
        Color of edge of nodes.  Only used if there is no node data.
    nodes_as_voronoi=False : bool
        If True, node data is visualized as colors of faces of a
        voronoi diagram based on node coordinates rather than color
        of circles at node coordinates.
    node_alpha=1 : float
        Transparency of nodes.
    node_quantity_cmap=None : colormap or None
        Colormap for node_data.
    node_quantity_clim=None : (float, float) or None
        Color limits for node_data.
    node_quantity_alpha=1 : float
        Transparency of colors used to represent node_data.
    node_quantity_logarithmic_colors=False : bool
        If True, node_data color-scale is logarithmic.
    face_quantity_cmap=None : colormap or None
        Colormap for face_data.
    face_quantity_clim=None : (float, float) or None
        Color limits for face_data.
    face_quantity_alpha=1 : float
        Transparency of colors used to represent face_data.
    face_quantity_logarithmic_colors=False : bool
        If True, face_data color-scale is logarithmic.
    figsize=None : (float, float) or None
        Size of figure in inches.
    title="" : str
        Title given to figure.

    """

    def __init__(self, config: TimeEvolutionResult, problem_nr=0, time_points=None, node_quantity=None,
                 junction_quantity="I", face_quantity=None, vortex_quantity="n",
                 vortex_diameter=0.25, vortex_color=(0, 0, 0), anti_vortex_color=(0.8, 0.1, 0.2),
                 vortex_alpha=1, _vortex_range=None,
                 show_grid=True, grid_width=1, grid_color=(0.4, 0.5, 0.6), grid_alpha=0.5,
                 show_colorbar=True, show_legend=True, show_axes=True,
                 arrow_width=0.005, arrow_scale=1, arrow_headwidth=3, arrow_headlength=5,
                 arrow_headaxislength=4.5, arrow_minshaft=1, arrow_minlength=1,
                 arrow_color=(0.15, 0.3, 0.8), arrow_alpha=1,
                 show_nodes=True, node_diameter=0.25, node_face_color=(1,1,1),
                 node_edge_color=(0, 0, 0),  nodes_as_voronoi=False,
                 node_alpha=1, node_quantity_cmap=None,
                 node_quantity_clim=None, node_quantity_alpha=1,
                 node_quantity_logarithmic_colors=False,
                 face_quantity_cmap=None, face_quantity_clim=None, face_quantity_alpha=1,
                 face_quantity_logarithmic_colors=False,
                 figsize=None, title="", animate_interval=10, fig=None):

        self.config = config
        self.problem_nr = problem_nr
        self.time_points = time_points
        if self.time_points is None:
            self.time_points = np.ones(self.config.problem._Nt(), dtype=bool)
        if not (self.time_points.dtype in (bool, np.bool)):
            try:
                self.time_points = np.zeros(self.config.problem._Nt(), dtype=bool)
                self.time_points[time_points] = True
            except:
                raise ValueError("Invalid time_points; must be None, mask, slice or index array")
        self.time_points &= self.config.problem.store_time_steps

        def _get_quantity(quantity, getter_func, problem_nr):
            """
            quantity -> "", "I", ndarray
            """
            if quantity == "" or quantity is None:
                return "none", None
            elif isinstance(quantity, str):
                all_data, quantity = getter_func(quantity)
            else:
                all_data, quantity = quantity, "custom"
            return quantity, all_data[:, problem_nr, :]

        self.node_quantity, node_data = _get_quantity(node_quantity, self._get_node_quantity, problem_nr)
        self.arrow_quantity, arrow_data = _get_quantity(junction_quantity, self._get_junction_quantity, problem_nr)
        self.face_quantity, face_data = _get_quantity(face_quantity, self._get_face_quantity, problem_nr)
        self.vortex_quantity, vortex_data = _get_quantity(vortex_quantity, self._get_vortex_quantity, problem_nr)

        if self.node_quantity == "phi":
            if node_quantity_clim is None:
                node_quantity_clim = (-np.pi, np.pi)

        if self.arrow_quantity == "th":
            if arrow_scale is None:
                arrow_scale = 1/np.pi

        super().__init__(config.get_circuit(),  node_data=node_data, arrow_data=arrow_data,
                         face_data=face_data, vortex_data=vortex_data,
                         vortex_label=self.vortex_quantity, node_label=self.node_quantity,
                         arrow_label=self.arrow_quantity, face_label=self.face_quantity,
                         vortex_diameter=vortex_diameter, vortex_color=vortex_color,
                         anti_vortex_color=anti_vortex_color,
                         vortex_alpha=vortex_alpha, show_grid=show_grid, grid_width=grid_width,
                         grid_color=grid_color,
                         grid_alpha=grid_alpha, show_colorbar=show_colorbar, show_legend=show_legend,
                         show_axes=show_axes,
                         arrow_width=arrow_width, arrow_scale=arrow_scale,
                         arrow_headwidth=arrow_headwidth, arrow_headlength=arrow_headlength,
                         arrow_headaxislength=arrow_headaxislength, arrow_minshaft=arrow_minshaft,
                         arrow_minlength=arrow_minlength,
                         arrow_color=arrow_color, arrow_alpha=arrow_alpha, show_nodes=show_nodes,
                         node_diameter=node_diameter, node_face_color=node_face_color,
                         node_edge_color=node_edge_color, nodes_as_voronoi=nodes_as_voronoi,
                         node_alpha=node_alpha,
                         node_quantity_cmap=node_quantity_cmap, node_quantity_clim=node_quantity_clim,
                         node_quantity_alpha=node_quantity_alpha,
                         node_quantity_logarithmic_colors=node_quantity_logarithmic_colors,
                         face_quantity_cmap=face_quantity_cmap,
                         face_quantity_clim=face_quantity_clim, face_quantity_alpha=face_quantity_alpha,
                         face_quantity_logarithmic_colors=face_quantity_logarithmic_colors,
                         figsize=figsize, title=title, animate_interval=animate_interval, fig=fig)


    _node_quantities = {
        "phi": 0, "phase": 0, "phases": 0,
        "I_s": 1, "Is": 1, "current_sources": 1,
        "U": 2, "potential": 2,
    }

    _junction_quantities = {
        "th": 0, "theta": 0, "phase_difference": 0, "gauge_invariant_phase_difference": 0,
        "I": 1, "current": 1,
        "V": 2, "voltage": 2,
        "Isup": 3, "I_sup": 3, "Isuper": 3, "I_super": 3, "supercurrent": 3, "super_current": 3,
        "I_s": 4, "Is": 4, "current_sources": 4,
        "EJ": 5, "josephson_energy": 5,
        "EM": 6, "magnetic_energy": 6,
        "EC": 7, "capacitive_energy": 7, "capacitance_energy": 7,
        "Etot": 8, "E_tot": 8, "ETot": 8, "total_energy": 8, "energy": 8,
    }

    _face_quantities = {
        "Phi": 0, "flux": 0, "magnetic_flux": 0,
        "n": 1, "vortices": 1, "vortex_configuration": 1,
        "face_current": 2, "J": 2,
    }

    _vortex_quantities = {
        "n": 0, "vortices": 0, "vortex_configuration": 0,
    }

    def get_node_quantity(self):
        """
        Get physical quantity displayed at nodes. (None, "phi", "Is_node" or "U")
        """
        return self.node_label

    def get_junction_quantity(self):
        """
        Get physical quantity displayed at junctions. (None, "th", "I",
        "V", "Isup", "Is", "EJ", "EM", "EC" or "Etot")
        """
        return self.arrow_label

    def get_face(self):
        """
        Get physical quantity displayed at faces with colors. (None, "flux", " n" or "J")
        """
        return self.face_label

    def get_vortex_quantity(self):
        """
        Get physical quantity displayed at faces with symbols. (None or "n")
        """
        return self.vortex_label


    def _get_node_quantity(self, node_quantity):
        """
        Get node quantity (either "phi": gauge dependent phases or  "U": potential)
        """
        quantity = self._node_quantities[node_quantity]
        if quantity == 0:  # phi
            out = self.config.get_phi(self.time_points)
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out, "phi"
        if quantity == 1:   # Is
            return self.config.problem.get_node_current_sources(self.time_points), "Is_node"
        if quantity == 2:   # U
            return self.config.get_U(self.time_points), "U"

    def _get_junction_quantity(self, junction_quantity):
        quantity = self._junction_quantities[junction_quantity]
        if quantity == 0:  # theta
            out = self.config.get_theta(self.time_points)
            out = out.copy()
            out -= np.round(out / (np.pi * 2.0)).astype(out.dtype) * np.pi * 2.0
            return out, "th"
        if quantity == 1:  # I
            return self.config.get_I(self.time_points), "I"
        if quantity == 2:  # V
            return self.config.get_V(self.time_points), "V"
        if quantity == 3:  # supercurrent
            return self.config.get_Isup(self.time_points), "Isup"
        if quantity == 4:  # Is
            return self.config.problem._Is(self.time_points), "Is"
        if quantity == 5:  # EJ
            return self.config.get_EJ(self.time_points), "EJ"
        if quantity == 6:  # EM
            return self.config.get_EM(self.time_points), "EM"
        if quantity == 7:  # EC
            return self.config.get_EC(self.time_points), "EC"
        if quantity == 8:  # Etot
            return self.config.get_Etot(self.time_points), "Etot"

    def _get_face_quantity(self, face_quantity):
        quantity = self._face_quantities[face_quantity]
        if quantity == 0:  # Phi
            return self.config.get_flux(self.time_points), "flux"
        if quantity == 1:  # n
            return self.config.get_n(self.time_points), "n"
        if quantity == 2:  # J
            return self.config.get_J(self.time_points), "J"

    def _get_vortex_quantity(self, vortex_quantity):
        quantity = self._vortex_quantities[vortex_quantity]
        if quantity == 0:  # n
            return self.config.get_n(self.time_points), "n"

